"""Definition of the SVG `<color>` type.

Use `Color` to represent colors in SVG. Use `ColorType` in Pydantic fields.
"""

import pydantic_extra_types.color
from typing_extensions import TypeAlias, override

from svglab import protocols, serialize
from svglab.utils import miscutils


def _alpha_channel_as_percentage(original: str) -> str:
    """Convert the alpha channel to a percentage.

    Args:
        original: The original color string. Must be in the format
            `func(r, g, b, a)`, where `a` is a float.

    Returns:
        The color string with the alpha channel as a percentage.

    Examples:
        >>> _alpha_channel_as_percentage("rgba(0, 0, 0, 0.5)")
        'rgba(0, 0, 0, 50%)'
        >>> _alpha_channel_as_percentage("hsla(0, 0%, 0%, 0.75)")
        'hsla(0, 0%, 0%, 75%)'

    """
    name_and_args = miscutils.extract_function_name_and_args(original)

    if name_and_args is None:
        msg = f"Invalid color string: {original!r}"
        raise ValueError(msg)

    func, args = name_and_args

    values = args.split(", ")
    values[-1] = f"{float(values[-1]):0.0%}"

    return f"{func}({', '.join(values)})"


class Color(
    pydantic_extra_types.color.Color, protocols.CustomSerializable
):
    """Represents the SVG `<color>` type.

    This class is used to represent colors in SVG.
    """

    @override
    def as_rgb(
        self, *, alpha_channel: serialize.AlphaChannelMode = "float"
    ) -> str:
        result = super().as_rgb()

        if self._rgba.alpha is not None and alpha_channel == "percentage":
            result = _alpha_channel_as_percentage(result)

        return result

    @override
    def as_hsl(
        self, *, alpha_channel: serialize.AlphaChannelMode = "float"
    ) -> str:
        result = super().as_hsl()

        if self._rgba.alpha is not None and alpha_channel == "percentage":
            result = _alpha_channel_as_percentage(result)

        return result

    @override
    def serialize(self) -> str:
        formatter = serialize.get_current_formatter()
        result: str

        match formatter.color_mode:
            case "original":
                original = self.original()

                if isinstance(original, str):
                    result = original
                else:
                    result = self.as_named(fallback=True)
            case "auto":
                result = super().__str__()
            case "named":
                result = self.as_named(fallback=True)
            case "hex-short":
                result = self.as_hex(format="short")
            case "hex-long":
                result = self.as_hex(format="long")
            case "rgb":
                result = self.as_rgb(alpha_channel=formatter.alpha_channel)
            case "hsl":
                result = self.as_hsl(alpha_channel=formatter.alpha_channel)

        if result.endswith(")"):  # just to make this a bit faster
            name_and_args = miscutils.extract_function_name_and_args(
                result
            )
            assert name_and_args is not None

            func, args_str = name_and_args
            args = args_str.split(", ")

            result = serialize.serialize_function_call(func, *args)

        return result

    @override
    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.original()!r})"


ColorType: TypeAlias = Color



"""Definition of the SVG `<length>` type.

Use `Length` to represent lengths in SVG. Use `LengthType` in Pydantic fields.
"""

from __future__ import annotations

import contextlib
from collections.abc import Iterator

import lark
import more_itertools
import pydantic
from typing_extensions import (
    Annotated,
    Final,
    Literal,
    Self,
    SupportsFloat,
    TypeAlias,
    final,
    override,
)

from svglab import errors, mixins, protocols, serialize, units, utiltypes
from svglab.attrparse import parse


_convert: Final[units.Converter[Length, utiltypes.LengthUnit]] = (
    units.make_converter(
        conversion_table={
            ("cm", "in"): 2.54,
            ("cm", "mm"): 10,
            ("pc", "px"): 15,
            ("pt", "px"): 1.25,
            (None, "px"): 1,
            ("mm", "Q"): 4,
        }
    )
)


@final
@pydantic.dataclasses.dataclass(frozen=True)
class Length(
    mixins.AddSub["Length"],
    mixins.FloatMulDiv,
    SupportsFloat,
    protocols.CustomSerializable,
):
    """Represents the SVG `<length>` type.

    A length is a number optionally followed by a unit. Available units are:
    - `%`: percentage
    - `ch`: character unit
    - `cm`: centimeters
    - `em`: relative to the font size of the element
    - `ex`: relative to the x-height of the element's font
    - `in`: inches
    - `mm`: millimeters
    - `pc`: picas
    - `pt`: points
    - `px`: pixels
    - `Q`: quarter-millimeters
    - `rem`: relative to the font size of the root element
    - `vh`: viewport height
    - `vmax`: maximum of the viewport's height and width
    - `vmin`: minimum of the viewport's height and width
    - `vw`: viewport width

    If the unit is set to `None`, the length is considered to be in user units.

    """

    value: float
    unit: utiltypes.LengthUnit = None

    def to(self, unit: utiltypes.LengthUnit) -> Length:
        """Convert the length to a different unit.

        Args:
            unit: The unit to convert to.

        Returns:
            A new `Length` object with the converted value and new unit.

        Raises:
            SvgUnitConversionError: If the conversion is not possible.

        Examples:
            >>> length = Length(10, "cm")
            >>> length.to("mm")
            Length(value=100.0, unit='mm')

        """
        return _convert(self, unit)

    @override
    def serialize(self) -> str:
        formatter = serialize.get_current_formatter()
        units: Iterator[utiltypes.LengthUnit | Literal["preserve"]] = (
            more_itertools.always_iterable(formatter.length_unit)
        )
        converted = self

        for unit in units:
            if unit == "preserve":
                break

            with contextlib.suppress(errors.SvgUnitConversionError):
                converted = self.to(unit)
                break

        value = serialize.serialize(
            converted.value, precision_group="coordinate"
        )

        return f"{value}{converted.unit or ''}"

    @classmethod
    def zero(cls) -> Length:
        """Return a length of zero."""
        return cls(0)

    @override
    def __add__(self, other: Length) -> Self:
        other_value = other.to(self.unit).value

        return type(self)(value=self.value + other_value, unit=self.unit)

    @override
    def __mul__(self, other: float) -> Self:
        return type(self)(value=self.value * other, unit=self.unit)

    def __bool__(self) -> bool:
        return bool(self.value)

    @override
    def __float__(self) -> float:
        return self.to(None).value


@lark.v_args(inline=True)
class _Transformer(lark.Transformer[object, Length]):
    number = float
    length = Length


LengthType: TypeAlias = Annotated[
    Length,
    parse.get_validator(grammar="length.lark", transformer=_Transformer()),
]



"""Utilities for parsing with Lark."""

import functools
import itertools
import os
import pathlib

import lark
import pydantic
from typing_extensions import Any, Final, LiteralString, TypeVar, cast


_T = TypeVar("_T")
_LeafT = TypeVar("_LeafT")
_ReturnT = TypeVar("_ReturnT")
_TransformerT = TypeVar("_TransformerT", bound=lark.Transformer[Any, Any])


_CURRENT_DIR: Final = pathlib.Path(__file__).parent
_GRAMMARS_DIR: Final = _CURRENT_DIR / "grammars"


def _get_grammar_path(grammar: LiteralString) -> pathlib.Path:
    """Get the path to a grammar file.

    Args:
        grammar: The name of the grammar file.

    Returns:
        The path to the grammar file.

    """
    return _GRAMMARS_DIR / grammar


@functools.cache
def _get_parser(*, grammar: LiteralString) -> lark.Lark:
    return lark.Lark.open(
        grammar_filename=str(_get_grammar_path(grammar)),
        rel_to=None,
        cache=True,
        maybe_placeholders=False,
        ordered_sets=False,
        parser="lalr",
        propagate_positions=False,
    )


def parse(
    text: str,
    /,
    *,
    grammar: LiteralString,
    transformer: lark.Transformer[_LeafT, _ReturnT],
) -> _ReturnT:
    """Parse a string using a Lark grammar.

    Args:
        text: The text to parse.
        grammar: The name of the grammar file.
        transformer: A Lark transformer to use for transforming the parse tree.
            The transformer must be a subclass of `lark.Transformer`.

    Returns:
        The parsed and transformed result.

    Raises:
        ValueError: If a parsing error occurs.

    """
    parser = _get_parser(grammar=grammar)

    try:
        tree = parser.parse(text)
        return transformer.transform(cast(lark.Tree[_LeafT], tree))
    except lark.LarkError as e:
        msg = f"Failed to parse text with grammar {grammar!r}. Reason: \n\n{e}"
        raise ValueError(msg) from e


def get_validator(
    *,
    grammar: LiteralString,
    transformer: lark.Transformer[_LeafT, _ReturnT],
    **kwargs: object,
) -> pydantic.BeforeValidator:
    """Get a Pydantic BeforeValidator for parsing with Lark.

    Args:
        grammar: The name of the grammar file.
        transformer: A Lark transformer to use for transforming the parse tree.
            The transformer must be a subclass of `lark.Transformer`.
        **kwargs: Additional keyword arguments to pass to the parser.

    Returns:
        A Pydantic BeforeValidator that can be used to parse strings with the
        specified grammar.

    Raises:
        ValueError: If the grammar file cannot be read.

    """
    grammar_path = _get_grammar_path(grammar)

    if not os.access(grammar_path, os.R_OK):
        msg = f"Cannot read grammar file: {grammar_path}"
        raise ValueError(msg)

    def validator(value: object) -> object:
        if isinstance(value, str):
            return parse(
                value, grammar=grammar, transformer=transformer, **kwargs
            )

        return value

    return pydantic.BeforeValidator(validator)


def v_args_to_list(*values: _T) -> list[_T]:
    """Convert the arguments to a list, discarding the first one.

    Args:
        *values: The values to convert.

    Returns:
        A list of the values, excluding the first one.

    """
    # drop the first value, which is the transformer instance (self)
    return list(itertools.islice(values, 1, None))


def visit_tokens(cls: type[_TransformerT]) -> type[_TransformerT]:
    """A decorator to make a Lark transformer visit tokens.

    This decorator modifies the `__init__` method of the transformer class
    to set the `visit_tokens` attribute to `True`.

    Args:
        cls: The transformer class to decorate.

    Returns:
        The decorated transformer class.

    """  # noqa: D401
    original_init = cls.__init__

    @functools.wraps(original_init)
    def new_init(
        self: _TransformerT, *args: object, **kwargs: object
    ) -> None:
        del args, kwargs
        original_init(self, visit_tokens=True)

    cls.__init__ = new_init
    return cls



"""Miscellaneous useful utility mixins.

This module is used to define mixins that can be used to add common
functionality to classes. The mixins are used to define common functionality
that is not specific to a particular class, but is useful for multiple classes.
"""

import abc

import pydantic
import pydantic_core
from typing_extensions import Any, Literal, Self, TypeVar, override

from svglab import protocols


_T_contra = TypeVar("_T_contra", contravariant=True)
_SupportsNegT_contra = TypeVar(
    "_SupportsNegT_contra", bound=protocols.SupportsNeg, contravariant=True
)
_SupportsRTrueDivT_contra = TypeVar(
    "_SupportsRTrueDivT_contra",
    bound=protocols.SupportsRTrueDiv[Any],
    contravariant=True,
)


class CustomModel(protocols.PydanticCompatible, metaclass=abc.ABCMeta):
    """A mixin for easy creation of custom pydantic-compatible classes.

    This class is a mixin for classes that need to be pydantic-compatible,
    but are not, for technical reasons, pydantic dataclasses or subclasses
    of `BaseModel` (for example collections).

    By implementing the `_validate` class method, the class can be used
    in pydantic models. `__get_pydantic_core_schema__` is created
    automatically and should not be overridden.
    """

    @classmethod
    @abc.abstractmethod
    def _validate(
        cls, value: object, info: pydantic_core.core_schema.ValidationInfo
    ) -> Self:
        """Validate the value and return a new instance of the class."""

    @override
    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: type, handler: pydantic.GetCoreSchemaHandler
    ) -> pydantic_core.core_schema.CoreSchema:
        del source_type, handler

        return (
            pydantic_core.core_schema.with_info_plain_validator_function(
                function=cls._validate
            )
        )


class SubWithNeg(
    protocols.SupportsAdd[_SupportsNegT_contra],
    protocols.SupportsSub[_SupportsNegT_contra],
    metaclass=abc.ABCMeta,
):
    """Implement subtraction using negation and addition."""

    @override
    def __sub__(self, other: _SupportsNegT_contra, /) -> Self:
        return self + -other


class RAddWithAdd(
    protocols.SupportsRAdd[_T_contra],
    protocols.SupportsAdd[_T_contra],
    metaclass=abc.ABCMeta,
):
    """Implement right addition using addition."""

    @override
    def __radd__(self, other: _T_contra, /) -> Self:
        return self + other


class RSubWithSub(
    protocols.SupportsRSub[_T_contra],
    protocols.SupportsSub[_T_contra],
    metaclass=abc.ABCMeta,
):
    """Implement right subtraction using subtraction."""

    @override
    def __rsub__(self, other: _T_contra, /) -> Self:
        return self - other


class RMulWithMul(
    protocols.SupportsRMul[_T_contra],
    protocols.SupportsMul[_T_contra],
    metaclass=abc.ABCMeta,
):
    """Implement right multiplication using multiplication."""

    @override
    def __rmul__(self, other: _T_contra, /) -> Self:
        return self * other


class RTruedivWithTrueDiv(
    protocols.SupportsRTrueDiv[_T_contra],
    protocols.SupportsTrueDiv[_T_contra],
    metaclass=abc.ABCMeta,
):
    """Implement right true division using true division."""

    @override
    def __rtruediv__(self, other: _T_contra, /) -> Self:
        return self / other


class NegWithMul(
    protocols.SupportsNeg,
    protocols.SupportsMul[Literal[-1]],
    metaclass=abc.ABCMeta,
):
    """Implement negation using multiplication by -1."""

    @override
    def __neg__(self) -> Self:
        return self * -1


class TrueDivWithMul(
    protocols.SupportsTrueDiv[_SupportsRTrueDivT_contra],
    protocols.SupportsMul[_SupportsRTrueDivT_contra],
    metaclass=abc.ABCMeta,
):
    """Implement true division using multiplication by the reciprocal."""

    @override
    def __truediv__(self, other: _SupportsRTrueDivT_contra, /) -> Self:
        return self * (1 / other)


class FloatMulDiv(
    NegWithMul,
    RMulWithMul[float],
    RTruedivWithTrueDiv[float],
    TrueDivWithMul[float],
    metaclass=abc.ABCMeta,
):
    """Implement multiplication and division as effortlessly as possible."""


class AddSub(
    SubWithNeg[_SupportsNegT_contra],
    RAddWithAdd[_SupportsNegT_contra],
    RSubWithSub[_SupportsNegT_contra],
    metaclass=abc.ABCMeta,
):
    """Implement addition and subtraction as effortlessly as possible."""



"""Mathematical utility functions."""

import math

from typing_extensions import Literal, SupportsFloat, SupportsIndex
from useful_types import SupportsRichComparisonT

from svglab import constants


def is_close(
    a: SupportsFloat | SupportsIndex, b: SupportsFloat | SupportsIndex, /
) -> bool:
    """Check if two floating-point numbers are almost equal.

    Args:
        a: The first number to compare.
        b: The second number to compare.

    Returns:
        `True` if the two numbers are almost equal, `False` otherwise.

    Examples:
        import math
        >>> is_close(1.0, 1.0)
        True
        >>> is_close(1.0, 1.0 + 1e-20)
        True
        >>> is_close(1.0, 1.0 + 0.001)
        False
        >>> is_close(0, math.sin(math.pi))
        True

    """
    return math.isclose(
        a,
        b,
        rel_tol=constants.FLOAT_RELATIVE_TOLERANCE,
        abs_tol=constants.FLOAT_ABSOLUTE_TOLERANCE,
    )


def clamp(
    value: SupportsRichComparisonT,
    /,
    *,
    min_value: SupportsRichComparisonT,
    max_value: SupportsRichComparisonT,
) -> SupportsRichComparisonT:
    """Clamp a value between two bounds.

    Args:
        value: The value to clamp.
        min_value: The minimum value.
        max_value: The maximum value.

    Returns:
        The clamped value. If `value` is less than `min_value`, `min_value` is
        returned. If `value` is greater than `max_value`, `max_value` is
        returned. Otherwise, `value` is returned.

    Examples:
        >>> clamp(5, min_value=0, max_value=10)
        5
        >>> clamp(-5, min_value=0, max_value=10)
        0
        >>> clamp(15, min_value=0, max_value=10)
        10

    """
    return max(min(value, max_value), min_value)


def signum(x: float) -> Literal[-1, 0, 1]:
    """Compute the signum function `sgn(x)`.

    Args:
        x: The value to compute the signum of.

    Returns:
        -1 if the value is negative, 0 if the value is zero, and 1 if the value
        is positive.

    Examples:
        >>> signum(-5)
        -1
        >>> signum(0)
        0
        >>> signum(0.5)
        1

    """
    if x < 0:
        return -1

    return 1 if x > 0 else 0


def arccos(value: float) -> float:
    """Compute the arccosine of a value in degrees.

    This function is a wrapper around `math.acos` that returns "nice" values
    for common inputs. The output is in degrees.

    Args:
        value: The value to compute the arccosine of.

    Returns:
        The arccosine of the value in degrees.

    Examples:
        >>> arccos(1)
        0
        >>> arccos(0)
        90

    """
    match value:
        case 1:
            return 0
        case 0:
            return 90
        case _:
            return math.degrees(math.acos(value))


def degrees(radians: float) -> float:
    """Convert radians to degrees, returning a value in the range (-180, 180].

    Args:
        radians: The angle in radians.

    Returns:
        The angle in degrees in the range (-180, 180].

    Examples:
        >>> from math import pi
        >>> degrees(0)
        0.0
        >>> degrees(pi)
        180.0
        >>> degrees(2 * pi)
        0.0
        >>> degrees(-pi / 2)
        -90.0
        >>> degrees(-pi)
        180.0

    """
    deg = math.degrees(radians)
    normalized = ((deg + 180) % 360) - 180

    return 180.0 if is_close(normalized, -180) else normalized


def arctan(value: float) -> float:
    """Compute the arctangent of a value in degrees.

    This function is a wrapper around `math.atan` that returns "nice" values
    for common inputs. The output is in degrees.

    Args:
        value: The value to compute the arctangent of.

    Returns:
        The arctangent of the value in degrees.

    Examples:
        >>> arctan(0)
        0
        >>> arctan(1)
        45

    """
    match value:
        case 0:
            return 0
        case 1:
            return 45
        case -1:
            return -45
        case _:
            return math.degrees(math.atan(value))


def tan(degrees: float) -> float:
    """Compute the tangent of an angle in degrees.

    This function is a wrapper around `math.tan` that takes an angle in degrees
    and returns "nice" values for common angles.

    Args:
        degrees: The angle in degrees.

    Returns:
        The tangent of the angle.

    Examples:
        >>> tan(0)
        0
        >>> tan(45)
        1

    """
    degrees %= 180

    match degrees:
        case 0:
            return 0
        case 45:
            return 1
        case 135:
            return -1
        case _:
            return math.tan(math.radians(degrees))



"""Definition of the SVG `<transform-list>` type and transformation functions.

Use `Transform` to represent transform lists in SVG.
Use `TransformType` in Pydantic fields.

Use `Translate`, `Scale`, `Rotate`, `SkewX`, `SkewY`, and `Matrix` to represent
transformations. Use `TransformFunction` to represent any transformation. Use
`Reifiable` to represent transformations that can be reified.
"""

from __future__ import annotations

import abc
import functools
import math
import operator
from collections.abc import Iterable
from types import NotImplementedType

import affine
import lark
import numpy.typing as npt
import pydantic
from typing_extensions import (
    Annotated,
    Self,
    TypeAlias,
    final,
    overload,
    override,
)

from svglab import mixins, protocols, serialize, utiltypes
from svglab.attrparse import parse
from svglab.utils import mathutils, miscutils


_Vector: TypeAlias = tuple[float, float]
"""A 2D vector."""


def _dot_product(u: _Vector, v: _Vector, /) -> float:
    """Calculate the dot product of two vectors.

    Args:
        u: The first vector.
        v: The second vector.

    Returns:
        The dot product of the two vectors.

    Examples:
        >>> _dot_product((1.0, 2.0), (3.0, 4.0))
        11.0
        >>> _dot_product((1.0, 2.0), (1.0, 2.0))
        5.0
        >>> _dot_product((1.0, 2.0), (0.0, 0.0))
        0.0

    """
    u1, u2 = u
    v1, v2 = v

    return u1 * v1 + u2 * v2


def _magnitude(u: _Vector, /) -> float:
    """Calculate the magnitude of a vector.

    Args:
        u: The vector.

    Returns:
        The magnitude of the vector.

    Examples:
        >>> _magnitude((3, 4))
        5.0
        >>> _magnitude((0, 0))
        0.0
        >>> _magnitude((1, 2))
        2.23606797749979

    """
    return math.sqrt(_dot_product(u, u))


class _TransformFunctionBase(
    protocols.CustomSerializable, metaclass=abc.ABCMeta
):
    @abc.abstractmethod
    def to_affine(self) -> affine.Affine:
        """Convert the transformation to an `affine.Affine` instance."""
        ...

    def to_matrix(self) -> Matrix:
        """Convert the transformation to a `Matrix` instance.

        Returns:
            The transformation as a matrix.

        Examples:
            >>> t = Translate(10, 20)
            >>> t.to_matrix()
            Matrix(a=1.0, b=0.0, c=0.0, d=1.0, e=10.0, f=20.0)
            >>> Matrix(1, 2, 3, 4, 5, 6).to_matrix()
            Matrix(a=1.0, b=2.0, c=3.0, d=4.0, e=5.0, f=6.0)

        """
        return Matrix.from_affine(self.to_affine())

    def __array__(
        self, dtype: npt.DTypeLike = None, *, copy: bool | None = None
    ) -> utiltypes.NpFloatArray:
        return self.to_affine().__array__(dtype=dtype, copy=copy)

    @overload
    def __matmul__(  # type: ignore[reportOverlappingOverload]
        self, other: _TransformFunctionBase
    ) -> Matrix: ...

    @overload
    def __matmul__(self, other: object) -> NotImplementedType: ...

    def __matmul__(self, other: object) -> Matrix | NotImplementedType:
        if not isinstance(other, _TransformFunctionBase):
            return NotImplemented

        product = self.to_affine() @ other.to_affine()
        assert isinstance(product, affine.Affine)

        return Matrix.from_affine(product)


@pydantic.dataclasses.dataclass(frozen=True)
class _Scale(_TransformFunctionBase):
    sx: float
    sy: float

    @override
    def serialize(self) -> str:
        args = [self.sx]

        if not mathutils.is_close(self.sx, self.sy):
            args.append(self.sy)

        return serialize.serialize_function_call(
            "scale", *args, precision_group="scale"
        )

    @override
    def to_affine(self) -> affine.Affine:
        return affine.Affine.scale(self.sx, self.sy or self.sx)

    def __eq__(self, other: object, /) -> bool:
        if not miscutils.basic_compare(other, self=self):
            return False

        return mathutils.is_close(
            self.sx, other.sx
        ) and mathutils.is_close(self.sy, other.sy)


@final
class Scale(_Scale):
    """A transformation that scales a shape by a given factor."""

    @overload
    def __init__(self, sx: float, /) -> None: ...

    @overload
    def __init__(self, sx: float, sy: float, /) -> None: ...

    def __init__(self, sx: float, sy: float | None = None, /) -> None:
        """Initialize a `Scale` transformation.

        Args:
            sx: The scaling factor on the x-axis.
            sy: The scaling factor on the y-axis. If not given, is assumed to
                be equal to `sx`.

        """
        super().__init__(sx, sy if sy is not None else sx)


@pydantic.dataclasses.dataclass(frozen=True)
class _Rotate(_TransformFunctionBase):
    angle: float
    cx: float
    cy: float

    @override
    def serialize(self) -> str:
        angle = serialize.serialize(self.angle, precision_group="angle")
        origin = []

        if not mathutils.is_close(self.cx, 0) or not mathutils.is_close(
            self.cy, 0
        ):
            origin.extend([self.cx, self.cy])

        return serialize.serialize_function_call("rotate", angle, *origin)

    @override
    def to_affine(self) -> affine.Affine:
        return affine.Affine.rotation(
            self.angle, (self.cx or 0, self.cy or 0)
        )

    def __eq__(self, other: object, /) -> bool:
        if not miscutils.basic_compare(other, self=self):
            return False

        return (
            mathutils.is_close(self.angle, other.angle)
            and mathutils.is_close(self.cx, other.cx)
            and mathutils.is_close(self.cy, other.cy)
        )


@final
class Rotate(_Rotate):
    """A transformation that rotates a shape by a given angle."""

    @overload
    def __init__(self, angle: float, /) -> None: ...

    @overload
    def __init__(self, angle: float, /, cx: float, cy: float) -> None: ...

    def __init__(
        self, angle: float, /, cx: float = 0, cy: float = 0
    ) -> None:
        """Initialize a `Rotate` transformation.

        Args:
            angle: The angle to rotate the shape by, in degrees. The rotation
                is counter-clockwise.
            cx: The x-coordinate of the point to rotate around. If not given,
                is assumed to be 0.
            cy: The y-coordinate of the point to rotate around. If not given,
                is assumed to be 0.

        """
        super().__init__(angle, cx, cy)


@final
@pydantic.dataclasses.dataclass(frozen=True)
class SkewY(_TransformFunctionBase):
    """A transformation that skews a shape along the y-axis.

    The shape is skewed by `angle` degrees along the y-axis.
    """

    angle: float

    @override
    def serialize(self) -> str:
        return serialize.serialize_function_call(
            "skewY", self.angle, precision_group="angle"
        )

    @override
    def to_affine(self) -> affine.Affine:
        return affine.Affine.shear(y_angle=self.angle)

    def __eq__(self, other: object, /) -> bool:
        if not miscutils.basic_compare(other, self=self):
            return False

        return mathutils.is_close(self.angle, other.angle)


@final
@pydantic.dataclasses.dataclass(frozen=True)
class SkewX(_TransformFunctionBase):
    """A transformation that skews a shape along the x-axis.

    The shape is skewed by `angle` degrees along the x-axis.
    """

    angle: float

    @override
    def serialize(self) -> str:
        return serialize.serialize_function_call(
            "skewX", self.angle, precision_group="angle"
        )

    @override
    def to_affine(self) -> affine.Affine:
        return affine.Affine.shear(x_angle=self.angle)

    def __eq__(self, other: object, /) -> bool:
        if not miscutils.basic_compare(other, self=self):
            return False

        return mathutils.is_close(self.angle, other.angle)


def _transform_weight(transform: Iterable[TransformFunction], /) -> int:
    """Calculate the weight of a transformation.

    The weight is used to determine the best decomposition of a matrix.

    Args:
        transform: The transformation to calculate the weight of.

    Returns:
        The weight of the transformation - the higher the weight, the more
        complex the transformation.

    """
    weight = 0

    for t in transform:
        match t:
            case SkewX() | SkewY():
                weight += 1_000_000
            case Scale(sx, sy) if not mathutils.is_close(sx, sy):
                weight += 1_000_000
            case Rotate():
                weight += 1_000
            case _:
                weight += 1

    return weight


@pydantic.dataclasses.dataclass(frozen=True)
class _Translate(_TransformFunctionBase):
    tx: float
    ty: float

    @override
    def serialize(self) -> str:
        args = [self.tx]

        if not mathutils.is_close(self.ty, 0):
            args.append(self.ty)

        return serialize.serialize_function_call("translate", *args)

    @override
    def to_affine(self) -> affine.Affine:
        return affine.Affine.translation(self.tx, self.ty or 0)

    def __eq__(self, other: object, /) -> bool:
        if not miscutils.basic_compare(other, self=self):
            return False

        return mathutils.is_close(
            self.tx, other.tx
        ) and mathutils.is_close(self.ty, other.ty)


@final
class Translate(_Translate):
    """A transformation that translates a shape by a given distance."""

    @overload
    def __init__(self, tx: float, /) -> None: ...

    @overload
    def __init__(self, tx: float, ty: float, /) -> None: ...

    def __init__(self, tx: float, ty: float = 0, /) -> None:
        """Initialize a `Translate` transformation.

        Args:
            tx: The distance to translate the shape by on the x-axis.
            ty: The distance to translate the shape by on the y-axis. If not
                given, is assumed to be equal to 0.

        """
        super().__init__(tx, ty)


def _remove_redundant_transformations(
    transform: Iterable[TransformFunction],
) -> Transform:
    result: Transform = []

    for t in transform:
        match t:
            case Translate(tx, ty) if mathutils.is_close(
                tx, 0
            ) and mathutils.is_close(ty, 0):
                continue
            case Scale(sx, sy) if mathutils.is_close(
                sx, 1
            ) and mathutils.is_close(sy, 1):
                continue
            case Rotate(angle) | SkewX(angle) | SkewY(angle) if (
                mathutils.is_close(angle, 0)
            ):
                continue
            case _:
                result.append(t)

    return result


@final
@pydantic.dataclasses.dataclass(frozen=True)
class Matrix(_TransformFunctionBase):
    """An arbitrary affine transformation.

    The transformation is represented by a 3x3 matrix in homogeneous
    coordinates:
    ```
    | a c e |
    | b d f |
    | 0 0 1 |
    ```
    """

    a: float
    b: float
    c: float
    d: float
    e: float
    f: float

    @classmethod
    def identity(cls) -> Self:
        """Create an identity matrix."""
        return cls(1, 0, 0, 1, 0, 0)

    @override
    def serialize(self) -> str:
        return serialize.serialize_function_call(
            "matrix", self.a, self.b, self.c, self.d, self.e, self.f
        )

    @override
    def to_affine(self) -> affine.Affine:
        # ! affine uses different labels for the matrix elements

        return affine.Affine(
            self.a, self.c, self.e, self.b, self.d, self.f
        )

    @classmethod
    def from_affine(cls, matrix: affine.Affine, /) -> Self:
        """Create a `Matrix` instance from an `affine.Affine` instance."""
        # ! affine uses different labels for the matrix elements

        return cls(
            matrix.a, matrix.d, matrix.b, matrix.e, matrix.c, matrix.f
        )

    def to_tuple(self) -> tuple[float, float, float, float, float, float]:
        """Convert the matrix to a tuple.

        Returns:
            The matrix as a tuple.

        Examples:
            >>> m = Matrix(1, 2, 3, 4, 5, 6)
            >>> m.to_tuple()
            (1.0, 2.0, 3.0, 4.0, 5.0, 6.0)

        """
        return self.a, self.b, self.c, self.d, self.e, self.f

    def determinant(self) -> float:
        """Calculate the determinant of the matrix.

        The determinant shows how the matrix scales the area of a shape.
        For example, a determinant of 1 means the area is unchanged.
        The determinant of 0 means the transformation is degenrate.

        Returns:
            The determinant of the matrix.

        Examples:
            >>> m = Matrix(1, 2, 3, 4, 5, 6)
            >>> m.determinant()
            -2.0

        """
        return self.to_affine().determinant

    def __qr_decompose(self) -> Transform:
        result: Transform = []
        a, b, c, d, e, f = self.to_tuple()

        # we prefer to use r over s if possible
        if not mathutils.is_close(a, 0) or not mathutils.is_close(b, 0):
            result.append(Translate(e, f))

            r = _magnitude((a, b))
            angle = mathutils.signum(b) * mathutils.arccos(a / r)
            result.append(Rotate(angle))

            det = self.determinant()
            result.append(Scale(r, det / r))

            col_dot = _dot_product((a, b), (c, d))
            angle = mathutils.arctan(col_dot / r**2)
            result.append(SkewX(angle))

        # if r is unsuitable, we use s
        elif not mathutils.is_close(c, 0) or not mathutils.is_close(d, 0):
            result.append(Translate(e, f))

            s = _magnitude((c, d))
            angle = 90 - mathutils.signum(d) * mathutils.arccos(-c / s)
            result.append(Rotate(angle))

            det = self.determinant()
            result.append(Scale(det / s, s))

            col_dot = _dot_product((a, b), (c, d))
            angle = mathutils.arctan(col_dot / s**2)
            result.append(SkewY(angle))

        # degenrate transformation
        else:
            result.append(Scale(0))

        return result

    def __ldu_decompose(self) -> Transform:
        result: Transform = []
        a, b, c, d, e, f = self.to_tuple()

        result.append(Translate(e, f))

        if not mathutils.is_close(a, 0):
            result.append(SkewY(mathutils.arctan(b / a)))
            result.append(Scale(a, self.determinant() / a))
            result.append(SkewX(mathutils.arctan(c / a)))
        elif not mathutils.is_close(b, 0):
            result.append(Rotate(90))
            result.append(Scale(b, self.determinant() / b))
            result.append(SkewY(mathutils.arctan(d / b)))
        else:
            result.append(Scale(c, d))
            result.append(SkewX(45))
            result.append(Scale(0, 1))

        return result

    def decompose(self) -> Transform:
        """Decompose the matrix into elementary transformations.

        The result is a list of transformations that, when composed, are
        equivalent to the original matrix.

        The algorithm is based on Frédéric Wang's [Decomposition of
        2D-transform matrices](https://frederic-wang.fr/2013/12/01/decomposition-of-2d-transform-matrices/).

        Two decomposition methods are used: QR and LDU. The
        final decomposition is chosen based on which method produces the
        transformation with the lowest complexity.

        Returns:
            A transformation list composed of elementary transformations.

        Examples:
        >>> m = Translate(10, 20).to_matrix()
        >>> m.decompose()
        [Translate(tx=10.0, ty=20.0)]
        >>> m = SkewY(45).to_matrix()
        >>> m.decompose()
        [SkewY(angle=45.0)]
        >>> m = Translate(10, 20) @ Scale(2, 2)
        >>> m.decompose()
        [Translate(tx=10.0, ty=20.0), Scale(sx=2.0, sy=2.0)]

        """
        decompositions = [self.__ldu_decompose(), self.__qr_decompose()]

        return min(
            map(_remove_redundant_transformations, decompositions),
            key=_transform_weight,
        )

    def __eq__(self, other: object, /) -> bool:
        if not miscutils.basic_compare(other, self=self):
            return False

        return all(
            mathutils.is_close(x1, x2)
            for x1, x2 in zip(
                self.to_tuple(), other.to_tuple(), strict=True
            )
        )


TransformFunction: TypeAlias = (
    Translate | Scale | Rotate | SkewX | SkewY | Matrix
)
"""A function that represents a transformation."""

Transform: TypeAlias = list[TransformFunction]
"""A list of transformations."""

Reifiable: TypeAlias = Translate | Scale
"""A transformation that can be reified."""


def decompose_matrices(transform: Transform) -> None:
    """Decompose matrices in a transformation list into elementary transforms.

    See `Matrix.decompose` for more information.

    Args:
        transform: The transformation list to decompose.

    Examples:
        >>> transform = [Matrix(1, 0, 0, 1, 10, 20)]
        >>> decompose_matrices(transform)
        >>> transform
        [Translate(tx=10.0, ty=20.0)]

    """
    i = 0

    while i < len(transform):
        transformation = transform[i]

        if not isinstance(transformation, Matrix):
            i += 1
            continue

        decomposition = transformation.decompose()
        transform[i : i + 1] = decomposition

        i += len(decomposition)


def compose(transforms: Iterable[TransformFunction], /) -> Matrix:
    """Compose a series of transformations into a single matrix.

    The transformations are applied in the order they are given.

    If no transformations are given, the identity matrix is returned.

    Args:
        transforms: The transformations to compose.

    Returns:
        The result of composing the transformations.

    Examples:
        >>> m1 = Matrix(1, 0, 0, 1, 2, 3)
        >>> m2 = Matrix(1, 0, 0, 1, 4, 5)
        >>> m3 = Matrix(1, 0, 0, 1, 6, 7)
        >>> compose([m1, m2, m3])
        Matrix(a=1.0, b=0.0, c=0.0, d=1.0, e=12.0, f=15.0)
        >>> compose([m1]) == m1
        True
        >>> compose([]) == Matrix.identity()
        True

    """
    return functools.reduce(operator.matmul, transforms, Matrix.identity())


class PointAddSubWithTranslateRMatmul(
    protocols.SupportsRMatmul["Translate"],
    mixins.AddSub[protocols.PointLike],
    metaclass=abc.ABCMeta,
):
    """Implement moving by a vector using multiplication with `Translate`."""

    @override
    def __add__(self, other: protocols.PointLike, /) -> Self:
        return Translate(other.x, other.y) @ self


@lark.v_args(inline=True)
class _Transformer(lark.Transformer[object, Transform]):
    number = float

    translate = Translate
    scale = Scale
    rotate = Rotate
    skew_x = SkewX
    skew_y = SkewY
    matrix = Matrix

    transform_ = parse.v_args_to_list


TransformType: TypeAlias = Annotated[
    Transform,
    parse.get_validator(
        grammar="transform.lark", transformer=_Transformer()
    ),
]



"""Utilities for working with iterables."""

import functools
from collections.abc import Generator, Iterable, Sequence

from typing_extensions import Sized, SupportsIndex, TypeVar


_T = TypeVar("_T")
_DT = TypeVar("_DT")


def take_last(iterable: Iterable[_T], /) -> _T | None:
    """Get the last item in an iterable.

    Args:
        iterable: The iterable to extract the last item from.

    Returns:
        The last item in the iterable, or `None` if the iterable is empty.

    Examples:
        >>> take_last([]) is None
        True
        >>> take_last([1, 2, 3])
        3
        >>> take_last(range(0)) is None
        True
        >>> take_last(range(3))
        2

    """
    return functools.reduce(lambda _, s: s, iterable, None)


def search_by_reference(sequence: Iterable[_T], item: _T) -> int:
    """Search for an item in a sequence by reference.

    Args:
        sequence: The sequence to search.
        item: The item to find.

    Returns:
        The index of the item in the sequence.

    Raises:
        ValueError: If the item is not found in the sequence.

    Examples:
        >>> search_by_reference([1, 2, 3], 2)
        1
        >>> search_by_reference([1, 2, 3], 4)
        Traceback (most recent call last):
            ...
        ValueError: Item not found in sequence: 4

    """
    for i, x in enumerate(sequence):
        if x is item:
            return i

    msg = f"Item not found in sequence: {item!r}"
    raise ValueError(msg)


def prev(sequence: Sequence[_T], item: _T) -> _T:
    """Get the item before a given item in a sequence.

    Args:
        sequence: The sequence to search.
        item: The item to find the predecessor of.

    Returns:
        The item before the given item in the sequence.

    Raises:
        ValueError: If the item is not found in the sequence or if the item


    Examples:
        >>> prev([1, 2, 3], 2)
        1
        >>> prev([1, 2, 3], 1)
        Traceback (most recent call last):
            ...
        ValueError: Item 1 has no predecessor.
        >>> prev([1, 2, 3], 4)
        Traceback (most recent call last):
            ...
        ValueError: Item not found in sequence: 4

    """
    idx = search_by_reference(sequence, item)

    if idx == 0:
        msg = f"Item {item!r} has no predecessor."
        raise ValueError(msg)

    return sequence[idx - 1]


def pairwise(
    iterable: Iterable[_T], /, default: _DT = None
) -> Generator[tuple[_T | _DT, _T]]:
    """Iterate over pairs of items in an iterable.

    Args:
        iterable: The iterable to iterate over.
        default: The default value to use for the first item.

    Yields:
        Pairs of items from the iterable. The first item in each pair is the
        previous item, or the default value if the first item is yielded.

    Examples:
        >>> list(pairwise([]))
        []
        >>> list(pairwise([1]))
        [(None, 1)]
        >>> list(pairwise([1, 2]))
        [(None, 1), (1, 2)]
        >>> list(pairwise([1, 2, 3]))
        [(None, 1), (1, 2), (2, 3)]

    """
    prev_item = default

    for item in iterable:
        yield prev_item, item
        prev_item = item


def is_first_index(sized: Sized, index: SupportsIndex) -> bool:
    """Check if an index resolves to the first index in a `Sized` object.

    Args:
        sized: The `Sized` object to check.
        index: The index to check.

    Returns:
        `True` if the index is the first index in the `Sized` object,
        `False` otherwise.

    Examples:
        >>> is_first_index([1, 2, 3], 0)
        True
        >>> is_first_index([1, 2, 3], 1)
        False
        >>> is_first_index([1, 2, 3], -1)
        False
        >>> is_first_index([1, 2, 3], -3)
        True
        >>> is_first_index([], 0)
        True

    """
    start, *_ = slice(index, index).indices(len(sized))

    return start == 0



"""Definition of the SVG `<path-data>` type.

Use `PathData` to represent path data in SVG. Use `PathDataType` in Pydantic
fields.
"""

from __future__ import annotations

import abc
import contextlib
from collections.abc import Generator, Iterable, MutableSequence

import lark
import pydantic
import pydantic_core
from typing_extensions import (
    Annotated,
    Final,
    Literal,
    Protocol,
    Self,
    SupportsIndex,
    TypeAlias,
    final,
    overload,
    override,
    runtime_checkable,
)

from svglab import errors, mixins, protocols, serialize
from svglab.attrparse import parse, point, transform
from svglab.utils import iterutils, miscutils


_Flag: TypeAlias = Literal["0", "1"]


@pydantic.dataclasses.dataclass
class _PathCommandBase:
    pass


@runtime_checkable
class _HasEnd(Protocol):
    end: point.Point


class _PhysicalPathCommand(
    _PathCommandBase,
    transform.PointAddSubWithTranslateRMatmul,
    metaclass=abc.ABCMeta,
):
    pass


@final
@pydantic.dataclasses.dataclass
class ClosePath(_PathCommandBase):
    """Close the current subpath (Z).

    The subpath is closed by drawing a straight line from the current
    point to current subpath's initial point.

    """


@pydantic.dataclasses.dataclass
class LineTo(_HasEnd, _PhysicalPathCommand):
    """Draw a line from the current point to the given end point (L)."""

    end: point.Point

    def __rmatmul__(self, other: transform.TransformFunction) -> Self:
        return type(self)(end=other @ self.end)


@final
@pydantic.dataclasses.dataclass
class HorizontalLineTo(_PhysicalPathCommand):
    """Draw a horizontal line from the current point (H)."""

    x: float

    def to_line(self) -> LineTo:
        """Convert to a `LineTo` instance."""
        return LineTo(end=point.Point(self.x, 0))

    @overload
    def __rmatmul__(
        self, other: transform.Translate | transform.Scale
    ) -> Self: ...

    @overload
    def __rmatmul__(self, other: transform.TransformFunction) -> Self: ...

    @override
    def __rmatmul__(
        self, other: transform.TransformFunction
    ) -> Self | LineTo:
        if isinstance(other, transform.Translate | transform.Scale):
            x, _ = other @ point.Point(self.x, 0)

            return type(self)(x=x)

        return other @ self.to_line()


@final
@pydantic.dataclasses.dataclass
class VerticalLineTo(_PhysicalPathCommand):
    """Draw a vertical line from the current point (V)."""

    y: float

    def to_line(self) -> LineTo:
        """Convert to a `LineTo` instance."""
        return LineTo(end=point.Point(0, self.y))

    @overload
    def __rmatmul__(
        self, other: transform.Translate | transform.Scale
    ) -> Self: ...

    @overload
    def __rmatmul__(self, other: transform.TransformFunction) -> Self: ...

    @override
    def __rmatmul__(
        self, other: transform.TransformFunction
    ) -> Self | LineTo:
        if isinstance(other, transform.Translate | transform.Scale):
            _, y = other @ point.Point(0, self.y)

            return type(self)(y=y)

        return other @ self.to_line()


@final
@pydantic.dataclasses.dataclass
class SmoothQuadraticBezierTo(_HasEnd, _PhysicalPathCommand):
    """Draw a smooth/shorthand quadratic Bézier curve (T).

    The curve is drawn from the current point to the end point. The control
    point is calculated based on the previous command as a reflection of the
    previous control point across the end point of the previous command.
    """

    end: point.Point

    def __rmatmul__(self, other: transform.TransformFunction) -> Self:
        return type(self)(end=other @ self.end)


@final
@pydantic.dataclasses.dataclass
class SmoothCubicBezierTo(_HasEnd, _PhysicalPathCommand):
    """Draw a smooth/shorthand cubic Bézier curve (S).

    The curve is drawn from the current point to the end point. The first
    control point is calculated based on the previous command as a reflection
    of the second control point of the previous command across the end point
    of the previous command. The second control point is given as an argument.
    """

    control2: point.Point
    end: point.Point

    def __rmatmul__(self, other: transform.TransformFunction) -> Self:
        return type(self)(
            control2=other @ self.control2, end=other @ self.end
        )


@final
@pydantic.dataclasses.dataclass
class MoveTo(_HasEnd, _PhysicalPathCommand):
    """Move the current point to the end point and start a new subpath (M)."""

    end: point.Point

    def __rmatmul__(self, other: transform.TransformFunction) -> Self:
        return type(self)(end=other @ self.end)


@final
@pydantic.dataclasses.dataclass
class QuadraticBezierTo(_HasEnd, _PhysicalPathCommand):
    """Draw a quadratic Bézier curve (Q).

    The curve is drawn from the current point to the end point using `control`
    as the control point.
    """

    control: point.Point
    end: point.Point

    def __rmatmul__(self, other: transform.TransformFunction) -> Self:
        return type(self)(
            control=other @ self.control, end=other @ self.end
        )


@final
@pydantic.dataclasses.dataclass
class CubicBezierTo(_HasEnd, _PhysicalPathCommand):
    """Draw a cubic Bézier curve (C).

    The curve is drawn from the current point to the end point using
    `control1` and `control2` as the control points.
    """

    control1: point.Point
    control2: point.Point
    end: point.Point

    def __rmatmul__(self, other: transform.TransformFunction) -> Self:
        return type(self)(
            control1=other @ self.control1,
            control2=other @ self.control2,
            end=other @ self.end,
        )


@final
@pydantic.dataclasses.dataclass
class ArcTo(_HasEnd, _PhysicalPathCommand):
    """Draw an elliptical arc (A).

    The arc is drawn from the current point to the end point using the
    following parameters:
    - `radii`: The radii of the arc.
    - `angle`: The x-axis rotation angle in degrees.
    - `large`: A flag indicating whether the arc is large or small.
    - `sweep`: A flag indicating whether the arc is drawn in a positive or
    negative angle direction.
    """

    radii: point.Point
    angle: float
    large: bool
    sweep: bool
    end: point.Point

    def __rmatmul__(self, other: transform.TransformFunction) -> Self:
        radii = self.radii
        angle = self.angle
        end = self.end

        match other:
            case transform.Translate():
                end = other @ end
            case transform.Scale():
                radii = other @ radii
                end = other @ end
            case transform.Rotate(a):
                angle += a
            case _:
                msg = f"Unsupported transform: {other}"
                raise NotImplementedError(msg)

        return type(self)(
            radii=radii,
            angle=angle,
            large=self.large,
            sweep=self.sweep,
            end=end,
        )


PathCommand: TypeAlias = (
    ArcTo
    | ClosePath
    | CubicBezierTo
    | HorizontalLineTo
    | LineTo
    | MoveTo
    | QuadraticBezierTo
    | SmoothCubicBezierTo
    | SmoothQuadraticBezierTo
    | VerticalLineTo
)


def _get_end(path_data: PathData, command: PathCommand) -> point.Point:
    """Get the end point of a path command.

    For commands that have a set end point, this function returns the end
    point. For commands that do not have a set end point, this function
    calculates the end point based on the previous command.

    Args:
        path_data: The path containing the command.
        command: The command to get the end point of.

    Returns:
        The end point of the command.

    Examples:
    >>> path_data = PathData.from_str("M 10,10 H 100 V 100 Z")
    >>> _get_end(path_data, path_data[0])
    Point(x=10.0, y=10.0)
    >>> _get_end(path_data, path_data[1])
    Point(x=100.0, y=10.0)

    """
    match command:
        case ClosePath():
            return _get_end(path_data, iterutils.prev(path_data, command))
        case HorizontalLineTo(x=x):
            end = _get_end(path_data, iterutils.prev(path_data, command))
            return point.Point(x, end.y)
        case VerticalLineTo(y=y):
            end = _get_end(path_data, iterutils.prev(path_data, command))
            return point.Point(end.x, y)
        case _:
            return command.end


def _quadratic_control(
    path_data: PathData, command: SmoothQuadraticBezierTo
) -> point.Point:
    """Compute the control point for a smooth quadratic Bézier command (`T`).

    The control point is calculated based on the previous command. If the
    previous command is a quadratic Bézier command (`Q`), the control point
    is the reflection of the previous control point across the end point of
    the previous command. If the previous command is of any other type, the
    control point is coincident with the end point of the previous command.

    Args:
        path_data: The path containing the command.
        command: The smooth quadratic Bézier command.

    Returns:
        The control point for the command.

    Examples:
    >>> path_data = PathData.from_str("M 0,0 Q 20,0 20,20 T 40,40")
    >>> _quadratic_control(path_data, path_data[2])
    Point(x=20.0, y=40.0)

    """
    prev = iterutils.prev(path_data, command)
    end = _get_end(path_data, prev)

    match prev:
        case QuadraticBezierTo(control=control):
            pass
        case SmoothQuadraticBezierTo():
            control = _quadratic_control(path_data, prev)
        case _:
            return end

    return control.line_reflect(end)


def _cubic_control(
    path_data: PathData, command: SmoothCubicBezierTo
) -> point.Point:
    """Compute the first control point for a smooth cubic Bézier command (`S`).

    The control point is calculated based on the previous command. If the
    previous command is a cubic Bézier command (`C`), the control point is the
    reflection of the second control point of the previous command across the
    end point of the previous command. If the previous command is of any other
    type, the control point is coincident with the end point of the previous
    command.

    Args:
        path_data: The path containing the command.
        command: The smooth cubic Bézier command.

    Returns:
        The first control point for the command.

    Examples:
    >>> path_data = PathData.from_str(
    ...     "M 0,0 C 20,0 20,20 40,40 S 100,100 50,50"
    ... )
    >>> _cubic_control(path_data, path_data[2])
    Point(x=60.0, y=60.0)

    """
    prev = iterutils.prev(path_data, command)
    end = _get_end(path_data, prev)

    if isinstance(prev, CubicBezierTo | SmoothCubicBezierTo):
        return prev.control2.line_reflect(end)

    return end


def _relativize(path_data: PathData) -> PathData:
    """Recompute the coordinates of path commands as if they were relative.

    This function takes a path and recomputes the coordinates of the path
    commands as if they were relative. This is done by subtracting the
    current position from the coordinates of each command. The resuting path
    may be used for serialization with relative coordinates.

    Args:
        path_data: The path to relativize.

    Returns:
        A new `PathData` instance with the coordinates of the path commands
        recomputed as if they were relative.

    Examples:
    >>> path_data = PathData.from_str("M 10,10")
    >>> _relativize(path_data)
    PathData(MoveTo(end=Point(x=10.0, y=10.0)))

    """
    result = PathData()
    pos = point.Point.zero()

    for command in path_data:
        if isinstance(command, _PhysicalPathCommand):
            result.append(command - pos)
        else:
            result.append(command)

        pos = _get_end(path_data, command)

    return result


def _can_use_implicit_command(
    current: PathCommand, /, *, prev: PathCommand | None
) -> bool:
    """Determine whether a command can be serialized implicitly.

    A command can be serialized implicitly if the previous command is of the
    same type as the current command, or if the previous command is a `MoveTo`
    command and the current command is a `LineTo` command.

    Args:
        prev: The previous command.
        current: The current command.

    Returns:
        `True` if the command can be serialized implicitly, `False` otherwise.

    Examples:
    >>> _can_use_implicit_command(MoveTo(point.Point(10, 10)), prev=None)
    False
    >>> _can_use_implicit_command(
    ...     LineTo(point.Point(100, 100)), prev=MoveTo(point.Point(10, 10))
    ... )
    True
    >>> _can_use_implicit_command(
    ...     LineTo(point.Point(100, 100)), prev=LineTo(point.Point(10, 10))
    ... )
    True

    """
    return type(prev) is type(current) or (
        isinstance(prev, MoveTo) and isinstance(current, LineTo)
    )


def _add_command(
    path_data: PathData, command: PathCommand, *, relative: bool = False
) -> None:
    if relative and isinstance(command, _PhysicalPathCommand):
        # if there is no previous command, just do nothing
        with contextlib.suppress(IndexError):
            command += _get_end(path_data, path_data[-1])

    path_data.append(command)


@final
class PathData(
    MutableSequence[PathCommand],
    mixins.CustomModel,
    transform.PointAddSubWithTranslateRMatmul,
    protocols.CustomSerializable,
):
    """A class representing the `d` attribute of a path element.

    The `d` attribute is used to define a path in SVG. This class provides
    methods for building a path by adding commands. `PathData` is also
    a `MutableSequence` of `PathCommand` instances.

    Args:
        iterable: An iterable of `PathCommand` instances (for example,
        another `PathData` instance).
        start: The starting point of the path. If `start` is not `None`, a
        `MoveTo` command is automatically added to the path, moving the "pen"
        to the starting point.

    Examples:
    >>> path_data = (
    ...     PathData()
    ...     .move_to(point.Point(10, 10))
    ...     .line_to(point.Point(100, 100), relative=True)
    ... )
    >>> len(path_data)
    2
    >>> bool(path_data)
    True
    >>> path_data[0]
    MoveTo(end=Point(x=10.0, y=10.0))
    >>> path_data.pop()
    LineTo(end=Point(x=110.0, y=110.0))
    >>> path_data.close()
    PathData(MoveTo(end=Point(x=10.0, y=10.0)), ClosePath())

    """

    def __add(
        self, command: PathCommand, /, *, relative: bool = False
    ) -> Self:
        _add_command(self, command, relative=relative)

        return self

    def __init__(
        self,
        iterable: Iterable[PathCommand] = (),
        /,
        *,
        start: point.Point | None = None,
    ) -> None:
        """Initialize a new `PathData` instance.

        Args:
            iterable: An iterable of `PathCommand` instances (for example,
            another `PathData` instance).
            start: The starting point of the path. If `start` is not `None`,
            a `MoveTo` command is automatically added to the path, moving the
            "pen" to the starting point.

        """
        self.__commands: Final[list[PathCommand]] = []

        if start is not None:
            self.move_to(start)

        for command in iterable:
            self.__add(command)

    @override
    def insert(self, index: SupportsIndex, value: PathCommand) -> None:
        if iterutils.is_first_index(self, index) and not isinstance(
            value, MoveTo
        ):
            raise errors.SvgPathMissingMoveToError

        self.__commands.insert(index, value)

    def move_to(
        self, end: point.Point, /, *, relative: bool = False
    ) -> Self:
        """Move the current point to the end point and start a new subpath (M).

        Args:
            end: The end point of the move command.
            relative: Whether the coordinates are relative to the current
            point.

        Returns:
            A reference to the current `PathData` instance.

        """
        return self.__add(MoveTo(end=end), relative=relative)

    def line_to(
        self, end: point.Point, /, *, relative: bool = False
    ) -> Self:
        """Draw a line from the current point to the end point (L).

        Args:
            end: The end point of the line command.
            relative: Whether the coordinates are relative to the current
            point.

        Returns:
            A reference to the current `PathData` instance.

        """
        return self.__add(LineTo(end=end), relative=relative)

    def horizontal_line_to(
        self,
        x: protocols.SupportsFloatOrIndex,
        /,
        *,
        relative: bool = False,
    ) -> Self:
        """Draw a horizontal line from the current point (H).

        Args:
            x: The x-coordinate of the end point of the line command.
            relative: Whether the coordinates are relative to the current
            point.

        Returns:
            A reference to the current `PathData` instance.

        """
        return self.__add(HorizontalLineTo(x=float(x)), relative=relative)

    def vertical_line_to(
        self,
        y: protocols.SupportsFloatOrIndex,
        /,
        *,
        relative: bool = False,
    ) -> Self:
        """Draw a vertical line from the current point (V).

        Args:
            y: The y-coordinate of the end point of the line command.
            relative: Whether the coordinates are relative to the current
            point.

        Returns:
            A reference to the current `PathData` instance.

        """
        return self.__add(VerticalLineTo(y=float(y)), relative=relative)

    def quadratic_bezier_to(
        self,
        control: point.Point,
        end: point.Point,
        /,
        *,
        relative: bool = False,
    ) -> Self:
        """Draw a quadratic Bézier curve (Q).

        The curve is drawn from the current point to the end point using
        `control` as the control point.

        Args:
            control: The control point of the curve.
            end: The end point of the curve.
            relative: Whether the coordinates are relative to the current
            point.

        Returns:
            A reference to the current `PathData` instance.

        """
        return self.__add(
            QuadraticBezierTo(control=control, end=end), relative=relative
        )

    def cubic_bezier_to(
        self,
        control1: point.Point,
        control2: point.Point,
        end: point.Point,
        *,
        relative: bool = False,
    ) -> Self:
        """Draw a cubic Bézier curve (C).

        The curve is drawn from the current point to the end point using
        `control1` and `control2` as the control points.

        Args:
            control1: The first control point of the curve.
            control2: The second control point of the curve.
            end: The end point of the curve.
            relative: Whether the coordinates are relative to the current
            point.

        Returns:
            A reference to the current `PathData` instance.

        """
        return self.__add(
            CubicBezierTo(control1=control1, control2=control2, end=end),
            relative=relative,
        )

    def arc_to(
        self,
        radii: point.Point,
        angle: float,
        end: point.Point,
        *,
        large: bool,
        sweep: bool,
        relative: bool = False,
    ) -> Self:
        """Draw an elliptical arc (A).

        The arc is drawn from the current point to the end point.

        Args:
            radii: The radii of the arc.
            angle: The x-axis rotation angle in degrees.
            end: The end point of the arc.
            large: A flag indicating whether the arc is large or small.
            sweep: A flag indicating whether the arc is drawn in a positive or
            negative angle direction.
            relative: Whether the coordinates are relative to the current
            point.

        Returns:
            A reference to the current `PathData` instance.

        """
        return self.__add(
            ArcTo(
                radii=radii, angle=angle, large=large, sweep=sweep, end=end
            ),
            relative=relative,
        )

    def smooth_quadratic_bezier_to(
        self, end: point.Point, /, *, relative: bool = False
    ) -> Self:
        """Draw a smooth/shorthand quadratic Bézier curve (T).

        The curve is drawn from the current point to the end point. The
        control point is calculated based on the previous command as a
        reflection of the previous control point across the end point of the
        previous command.

        Args:
            end: The end point of the curve.
            relative: Whether the coordinates are relative to the current
            point.

        Returns:
            A reference to the current `PathData` instance.

        """
        return self.__add(
            SmoothQuadraticBezierTo(end=end), relative=relative
        )

    def smooth_cubic_bezier_to(
        self,
        control2: point.Point,
        end: point.Point,
        /,
        *,
        relative: bool = False,
    ) -> Self:
        """Draw a smooth/shorthand cubic Bézier curve (S).

        The curve is drawn from the current point to the end point. The first
        control point is calculated based on the previous command as a
        reflection of the second control point of the previous command across
        the end point of the previous command. The second control point is
        given as an argument.

        Args:
            control2: The second control point of the curve.
            end: The end point of the curve.
            relative: Whether the coordinates are relative to the current
            point.

        Returns:
            A reference to the current `PathData` instance.

        """
        return self.__add(
            SmoothCubicBezierTo(control2=control2, end=end),
            relative=relative,
        )

    def close(self) -> Self:
        """Close the current subpath (Z).

        The subpath is closed by drawing a straight line from the current
        point to the current subpath's initial point.

        Returns:
            A reference to the current `PathData` instance.

        """
        return self.__add(ClosePath())

    @classmethod
    def from_str(cls, text: str) -> Self:
        """Parse a string into a `PathData` instance.

        Args:
            text: The string to parse.

        Returns:
            A new `PathData` instance created from the parsed string.

        Raises:
            ValueError: If the string cannot be parsed into a valid `PathData`
            instance.

        Examples:
            >>> path_data = PathData.from_str("M 10,10 Z")
            >>> path_data
            PathData(MoveTo(end=Point(x=10.0, y=10.0)), ClosePath())

        """
        path_data = parse.parse(
            text, grammar="path_data.lark", transformer=_Transformer()
        )

        assert isinstance(path_data, cls), (
            f"Expected {cls}, got {type(path_data)}"
        )
        return path_data

    @classmethod
    def _validate(
        cls, value: object, info: pydantic_core.core_schema.ValidationInfo
    ) -> PathData:
        del info

        match value:
            case str():
                return cls.from_str(value)
            case PathData():
                return value
            case _:
                msg = f"Expected str or D, got {type(value)}"
                raise TypeError(msg)

    def __apply_shorthand_formatting(self) -> PathData:
        """Apply shorthand formatting based on the formatter settings."""
        formatter = serialize.get_current_formatter()
        line = formatter.path_data_shorthand_line_commands
        curve = formatter.path_data_shorthand_curve_commands

        path_data = self

        if "never" in (line, curve):
            path_data = path_data.resolve_shorthands(
                lines=line == "never", curves=curve == "never"
            )

        if "always" in (line, curve):
            path_data = path_data.apply_shorthands(
                lines=line == "always", curves=curve == "always"
            )

        return path_data

    def __serialize_commands(self) -> Generator[str]:
        formatter = serialize.get_current_formatter()
        path_data = self.__apply_shorthand_formatting()

        if formatter.path_data_coordinates == "relative":
            path_data = _relativize(path_data)

        for prev, command in iterutils.pairwise(path_data):
            implicit = (
                formatter.path_data_commands == "implicit"
                and _can_use_implicit_command(command, prev=prev)
            )

            match command:
                case MoveTo(end):
                    yield serialize.serialize_path_command(
                        end, char="M", implicit=implicit
                    )
                case LineTo(end):
                    yield serialize.serialize_path_command(
                        end, char="L", implicit=implicit
                    )
                case HorizontalLineTo(x):
                    yield serialize.serialize_path_command(
                        x, char="H", implicit=implicit
                    )
                case VerticalLineTo(y):
                    yield serialize.serialize_path_command(
                        y, char="V", implicit=implicit
                    )
                case QuadraticBezierTo(control, end):
                    yield serialize.serialize_path_command(
                        control, end, char="Q", implicit=implicit
                    )
                case SmoothQuadraticBezierTo(end):
                    yield serialize.serialize_path_command(
                        end, char="T", implicit=implicit
                    )
                case CubicBezierTo(control1, control2, end):
                    yield serialize.serialize_path_command(
                        control1,
                        control2,
                        end,
                        char="C",
                        implicit=implicit,
                    )
                case SmoothCubicBezierTo(control2, end):
                    yield serialize.serialize_path_command(
                        control2, end, char="S", implicit=implicit
                    )
                case ArcTo(radii, angle, large, sweep, end):
                    yield serialize.serialize_path_command(
                        radii,
                        angle,
                        large,
                        sweep,
                        end,
                        char="A",
                        implicit=implicit,
                    )
                case ClosePath():
                    yield serialize.serialize_path_command(
                        char="Z", implicit=implicit
                    )

    @override
    def serialize(self) -> str:
        return " ".join(self.__serialize_commands())

    def resolve_shorthands(
        self, *, lines: bool = True, curves: bool = True
    ) -> Self:
        """Resolve shorthand commands into their full-length equivalents.

        Args:
            lines: Whether to resolve shorthand line commands (`H`, `V`) into
            the full-length equivalent (`L`).
            curves: Whether to resolve shorthand curve commands (`S`, `T`) into
            the full-length equivalent (`C`, `Q`).

        Returns:
            A new `PathData` instance with the shorthand commands
            (`H`, `V`, `S`, `T`) replaced by their full-length equivalents
            (`L`, `C`, `Q`).

        Examples:
        >>> path_data = PathData.from_str("M 0,0")
        >>> path_data.resolve_shorthands()
        PathData(MoveTo(end=Point(x=0.0, y=0.0)))

        """
        path_data = type(self)()

        for command in self:
            match command:
                case SmoothQuadraticBezierTo(end=end) if curves:
                    control = _quadratic_control(self, command)
                    path_data.quadratic_bezier_to(control, end)
                case SmoothCubicBezierTo(control2=control2, end=end) if (
                    curves
                ):
                    control1 = _cubic_control(self, command)
                    path_data.cubic_bezier_to(control1, control2, end)
                case HorizontalLineTo(x=x) if lines:
                    end = _get_end(path_data, path_data[-1])
                    path_data.line_to(point.Point(x, end.y))
                case VerticalLineTo(y=y) if lines:
                    end = _get_end(path_data, path_data[-1])
                    path_data.line_to(point.Point(end.x, y))
                case _:
                    path_data.append(command)
        return path_data

    def apply_shorthands(
        self, *, lines: bool = True, curves: bool = True
    ) -> Self:
        """Convert full-length commands into shorthand commands where possible.

        Args:
            lines: Whether to convert line commands (`L`) to shorthand
            (`H`, `V`) where possible.
            curves: Whether to convert curve commands (`Q`, `C`) to shorthand
            (`T`, `S`) where possible.

        Returns:
            A new `PathData` instance with the full-length commands
            (`L`, `C`, `Q`) replaced by their shorthand equivalents
            (`H`, `V`, `S`, `T`).

        Examples:
        >>> path_data = PathData.from_str("M 10,10 L 100,10")
        >>> path_data.apply_shorthands()
        PathData(MoveTo(end=Point(x=10.0, y=10.0)), HorizontalLineTo(x=100.0))

        """
        path_data = type(self)()

        for command in self:
            match command:
                case LineTo(end=end) if lines and end.x == (
                    _get_end(path_data, path_data[-1]).x
                ):
                    path_data.vertical_line_to(end.y)
                case LineTo(end=end) if lines and end.y == (
                    _get_end(path_data, path_data[-1]).y
                ):
                    path_data.horizontal_line_to(end.x)
                case QuadraticBezierTo(control=control, end=end) if curves:
                    path_data.smooth_quadratic_bezier_to(end)

                    shorthand = path_data[-1]
                    assert isinstance(shorthand, SmoothQuadraticBezierTo)

                    # try to replace the command with a shorthand
                    auto_control = _quadratic_control(path_data, shorthand)

                    # if the shorthand turns out not to be compatible, revert
                    # to the original command
                    if control != auto_control:
                        path_data[-1] = command

                case CubicBezierTo(
                    control1=control1, control2=control2, end=end
                ) if curves:
                    path_data.smooth_cubic_bezier_to(control2, end)

                    shorthand = path_data[-1]
                    assert isinstance(shorthand, SmoothCubicBezierTo)

                    auto_control = _cubic_control(path_data, shorthand)

                    if control1 != auto_control:
                        path_data[-1] = command

                case _:
                    path_data.append(command)

        return path_data

    @overload
    def __getitem__(self, index: SupportsIndex) -> PathCommand: ...

    @overload
    def __getitem__(self, index: slice) -> Self: ...

    @override
    def __getitem__(
        self, index: SupportsIndex | slice
    ) -> PathCommand | Self:
        if isinstance(index, SupportsIndex):
            return self.__commands[index]

        return type(self)(self.__commands[index])

    @overload
    def __setitem__(
        self, index: SupportsIndex, values: PathCommand
    ) -> None: ...

    @overload
    def __setitem__(
        self, index: slice, values: Iterable[PathCommand]
    ) -> None: ...

    @override
    def __setitem__(
        self,
        index: SupportsIndex | slice,
        values: PathCommand | Iterable[PathCommand],
    ) -> None:
        if isinstance(index, slice):
            raise NotImplementedError(
                "__setitem__ with a slice is not supported"
            )

        assert isinstance(values, PathCommand)

        if iterutils.is_first_index(self, index) and not isinstance(
            values, MoveTo
        ):
            raise errors.SvgPathMissingMoveToError

        self.__commands[index] = values

    @overload
    def __delitem__(self, index: SupportsIndex) -> None: ...

    @overload
    def __delitem__(self, index: slice) -> None: ...

    @override
    def __delitem__(self, index: SupportsIndex | slice) -> None:
        if isinstance(index, slice):
            raise NotImplementedError(
                "__delitem__ with a slice is not supported"
            )

        if (
            iterutils.is_first_index(self, index)
            and len(self) > 1
            and not isinstance(self[int(index) + 1], MoveTo)
        ):
            raise errors.SvgPathMissingMoveToError

        del self.__commands[index]

    @override
    def __len__(self) -> int:
        return len(self.__commands)

    @override
    def __rmatmul__(self, other: transform.TransformFunction) -> Self:
        return type(self)(
            other @ command
            if isinstance(command, _PhysicalPathCommand)
            else command
            for command in self
        )

    @override
    def __eq__(self, other: object) -> bool:
        if not miscutils.basic_compare(other, self=self):
            return False

        if len(self) != len(other):
            return False

        return all(c1 == c2 for c1, c2 in zip(self, other, strict=True))

    @override
    def __repr__(self) -> str:
        name = type(self).__name__
        commands = ", ".join(repr(command) for command in self)
        return f"{name}({commands})"


@lark.v_args(inline=True)
@parse.visit_tokens  # there are a few terminals we want to parse
class _Transformer(lark.Transformer[object, PathData]):
    point = point.Point
    NUMBER = float

    cubic_bezier = CubicBezierTo
    horizontal_line = HorizontalLineTo
    line = LineTo
    move = MoveTo
    quadratic_bezier = QuadraticBezierTo
    smooth_cubic_bezier = SmoothCubicBezierTo
    smooth_quadratic_bezier = SmoothQuadraticBezierTo
    vertical_line = VerticalLineTo
    z = ClosePath

    def FLAG(self, value: _Flag) -> bool:  # noqa: N802
        return value == "1"

    def arc(
        self,
        radii: point.Point,
        angle: float,
        large: bool,  # noqa: FBT001
        sweep: bool,  # noqa: FBT001
        end: point.Point,
    ) -> ArcTo:
        return ArcTo(
            radii=radii, angle=angle, large=large, sweep=sweep, end=end
        )

    @lark.v_args(inline=False)
    def path(
        self, args: list[PathCommand | lark.Tree[PathCommand | lark.Token]]
    ) -> PathData:
        path_data = PathData()

        for item in args:
            match item:
                # simple commands like `Z` require no further processing
                case _PathCommandBase() as command:
                    _add_command(path_data, command)
                # commands that are part of a group need to be extracted;
                # the relative flag is applied if the group is relative
                case lark.Tree(data=name, children=commands):
                    # sanity check for when the grammar changes
                    assert "relative" in name or "absolute" in name
                    relative = "relative" in name

                    for command in commands:
                        assert isinstance(command, PathCommand)
                        _add_command(path_data, command, relative=relative)

        return path_data


PathDataType: TypeAlias = Annotated[
    PathData,
    parse.get_validator(
        grammar="path_data.lark", transformer=_Transformer()
    ),
]



"""Business logic exceptions for the svglab package."""

from typing_extensions import override


class SvgError(Exception):
    """Base class for exceptions in this package."""


class SvgElementNotFoundError(SvgError):
    """Exception raised by methods that search the element tree."""


class SvgPathError(SvgError):
    """Exception raised by methods that manipulate paths."""


class SvgPathMissingMoveToError(SvgPathError):
    """Exception raised when a path does not start with a MoveTo command."""

    @override
    def __init__(self) -> None:
        super().__init__("Path must start with a MoveTo command")


class SvgUnitConversionError(SvgError):
    """Exception raised when a unit conversion fails."""

    @override
    def __init__(
        self, *, original_unit: object, target_unit: object
    ) -> None:
        super().__init__(
            f"Unable to convert {original_unit!r} to {target_unit!r}"
        )


class SvgTransformSwapError(SvgError):
    """Exception raised when two transformations cannot be swapped."""

    @override
    def __init__(self, transform_a: object, transform_b: object) -> None:
        super().__init__(
            f"Cannot swap {transform_a!r} and {transform_b!r}"
        )


class SvgTransformOriginError(SvgError):
    """Exception raised when a transform origin value is unsupported."""

    @override
    def __init__(self, transform_origin: object) -> None:
        super().__init__(
            f"Unsupported transform-origin: {transform_origin!r}"
        )



"""Useful constants that should be available globally."""

import typing_extensions
from typing_extensions import Final

from svglab import utiltypes


SVG_XMLNS: Final[utiltypes.Xmlns] = typing_extensions.get_args(
    utiltypes.Xmlns
)[0]
"""The value of the `xmlns` attribute."""

FLOAT_RELATIVE_TOLERANCE: Final = 1e-9
"""The relative tolerance used for floating-point comparisons."""

FLOAT_ABSOLUTE_TOLERANCE: Final = 1e-9
"""The absolute tolerance used for floating-point comparisons."""



"""Utility types that should be defined globally."""

import numpy as np
import numpy.typing as npt
from typing_extensions import Literal, TypeAlias


Xmlns: TypeAlias = Literal["http://www.w3.org/2000/svg"]
"""Type representing valid values for the `xmlns` attribute."""

NpFloat: TypeAlias = np.float64
"""The floating-point type used for numpy arrays."""

NpFloatArray: TypeAlias = npt.NDArray[NpFloat]
"""An ndarray of floating-point numbers."""

LengthUnit: TypeAlias = (
    Literal[
        "%",
        "ch",
        "cm",
        "em",
        "ex",
        "in",
        "mm",
        "pc",
        "pt",
        "px",
        "Q",
        "rem",
        "vh",
        "vmax",
        "vmin",
        "vw",
    ]
    | None
)
"""Units of `<length>`."""



"""Definitions and parsing of SVG attribute types.

This package provides classes to represent SVG attribute types that cannot be
parsed directly by Pydantic, such as `<transform-list>`, `<list-of-points>`,
`<color>`, and `<length>`.

It also provides logic to parse strings into these types and include them in
Pydantic models.
"""



"""Definitions of SVG elements.

This module contains the definitions of all SVG elements, including their
attributes and methods.

The elements are defined as classes, which inherit traits from the `traits`
module and attributes from the `attrs` package.
"""

# ruff: noqa: D101

from __future__ import annotations

import base64
import contextlib
import itertools
import os
import pathlib

import PIL.Image
from typing_extensions import final, overload, override

from svglab import graphics, models, protocols, serialize
from svglab.attrparse import length, path_data, point, transform
from svglab.attrs import attrdefs, attrgroups
from svglab.elements import traits
from svglab.utils import mathutils


# ! WARNING: `Element` and `ContainerElement` must always go last


def _length_or_zero(value: length.Length | None, /) -> length.Length:
    return value if value is not None else length.Length(0)


@final
class Path(
    attrgroups.ConditionalProcessingAttrs,
    attrdefs.ClassAttr,
    attrdefs.DAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.StyleAttr,
    traits.Shape,
    traits.Element,
):
    pass


def _basic_shape_to_path(basic_shape: traits.BasicShape, /) -> Path:
    """Convert a basic shape to a `Path` element."""
    # try to convert to D first, so we don't call convert() if the shape
    # is not convertible
    d = basic_shape.to_path_data()

    path = models.convert(basic_shape, Path)
    path.d = d

    return path


def _points_to_path_data(
    element: attrdefs.PointsAttr,
) -> path_data.PathData:
    """Convert element with a `points` attribute to a `PathData` instance."""
    points = element.points if element.points is not None else []
    d = path_data.PathData()

    if points:
        d.move_to(points[0])

        for point in itertools.islice(points, 1, None):
            d.line_to(point)

    return d


def _ellipse_to_path_data(
    *,
    cx: length.Length,
    cy: length.Length,
    rx: length.Length,
    ry: length.Length,
) -> path_data.PathData:
    """Convert an ellipse to a `PathData` instance.

    Args:
        cx: The x-coordinate of the center of the ellipse.
        cy: The y-coordinate of the center of the ellipse.
        rx: The x-radius of the ellipse.
        ry: The y-radius of the ellipse.

    Returns:
        A `PathData` instance representing the ellipse.

    References:
        https://stackoverflow.com/a/10477334
        https://aleen42.gitbooks.io/wiki/content/Programming/JavaScript/webgl/SVG/convert_shapes_to_path/convert_shapes_to_path.html

    """
    return (
        path_data.PathData()
        .move_to(point.Point(cx - rx, cy))
        .arc_to(
            point.Point(rx, ry),
            0,
            point.Point(2 * rx, 0),
            large=True,
            sweep=False,
            relative=True,
        )
        .arc_to(
            point.Point(rx, ry),
            0,
            point.Point(-2 * rx, 0),
            large=True,
            sweep=False,
            relative=True,
        )
    )


@final
class A(
    attrgroups.ConditionalProcessingAttrs,
    attrgroups.XlinkAttrs,
    attrdefs.ClassAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.StyleAttr,
    attrdefs.TargetAttr,
    traits.ContainerElement,
):
    pass


@final
class AltGlyph(
    attrgroups.ConditionalProcessingAttrs,
    attrgroups.XlinkAttrs,
    attrdefs.ClassAttr,
    attrdefs.DxListOfLengthsAttr,
    attrdefs.DyListOfLengthsAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.FormatAttr,
    attrdefs.GlyphRefAttr,
    attrdefs.RotateListOfNumbersAttr,
    attrdefs.StyleAttr,
    attrdefs.XListOfCoordinatesAttr,
    attrdefs.YListOfCoordinatesAttr,
    traits.TextContentChildElement,
    traits.TextContentElement,
    traits.Element,
):
    pass


@final
class AltGlyphDef(traits.Element):
    pass


@final
class AltGlyphItem(traits.Element):
    pass


@final
class Animate(
    attrgroups.AnimationAdditionAttrs,
    attrgroups.AnimationAttributeTargetAttrs,
    attrgroups.AnimationValueAttrs,
    attrgroups.ConditionalProcessingAttrs,
    attrgroups.XlinkAttrs,
    traits.AnimationElement,
    traits.Element,
):
    pass


@final
class AnimateColor(
    attrgroups.AnimationAdditionAttrs,
    attrgroups.AnimationAttributeTargetAttrs,
    attrgroups.AnimationValueAttrs,
    attrgroups.ConditionalProcessingAttrs,
    attrgroups.XlinkAttrs,
    attrdefs.ExternalResourcesRequiredAttr,
    traits.AnimationElement,
    traits.Element,
):
    pass


@final
class AnimateMotion(
    attrgroups.AnimationAdditionAttrs,
    attrgroups.AnimationValueAttrs,
    attrgroups.ConditionalProcessingAttrs,
    attrgroups.XlinkAttrs,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.KeyPointsAttr,
    attrdefs.OriginAttr,
    attrdefs.PathAttr,
    attrdefs.RotateNumberAutoAutoReverseAttr,
    traits.AnimationElement,
    traits.Element,
):
    pass


@final
class AnimateTransform(
    attrgroups.AnimationAdditionAttrs,
    attrgroups.AnimationAttributeTargetAttrs,
    attrgroups.AnimationValueAttrs,
    attrgroups.ConditionalProcessingAttrs,
    attrgroups.XlinkAttrs,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.TypeAnimateTransformAttr,
    traits.AnimationElement,
    traits.Element,
):
    pass


@final
class Circle(
    attrgroups.ConditionalProcessingAttrs,
    attrdefs.ClassAttr,
    attrdefs.CxAttr,
    attrdefs.CyAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.RAttr,
    attrdefs.StyleAttr,
    traits.BasicShape,
    traits.Element,
):
    @override
    def to_path_data(self) -> path_data.PathData:
        cx = _length_or_zero(self.cx)
        cy = _length_or_zero(self.cy)
        r = _length_or_zero(self.r)

        return _ellipse_to_path_data(cx=cx, cy=cy, rx=r, ry=r)

    @override
    def to_path(self) -> Path:
        return _basic_shape_to_path(self)


@final
class ClipPath(
    attrgroups.ConditionalProcessingAttrs,
    attrdefs.ClassAttr,
    attrdefs.ClipPathUnitsAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.StyleAttr,
    traits.Element,
):
    pass


@final
class ColorProfile(
    attrgroups.XlinkAttrs,
    attrdefs.LocalAttr,
    attrdefs.NameNameAttr,
    attrdefs.RenderingIntentAttr,
    traits.Element,
):
    pass


@final
class Cursor(
    attrgroups.ConditionalProcessingAttrs,
    attrgroups.XlinkAttrs,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.XCoordinateAttr,
    attrdefs.YCoordinateAttr,
    traits.Element,
):
    pass


@final
class Defs(
    attrgroups.ConditionalProcessingAttrs,
    attrdefs.ClassAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.StyleAttr,
    traits.StructuralElement,
    traits.ContainerElement,
):
    pass


@final
class Desc(
    attrdefs.ClassAttr,
    attrdefs.StyleAttr,
    traits.DescriptiveElement,
    traits.Element,
):
    pass


@final
class Ellipse(
    attrgroups.ConditionalProcessingAttrs,
    attrdefs.ClassAttr,
    attrdefs.CxAttr,
    attrdefs.CyAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.RxAttr,
    attrdefs.RyAttr,
    attrdefs.StyleAttr,
    traits.BasicShape,
    traits.Element,
):
    @override
    def to_path_data(self) -> path_data.PathData:
        cx = _length_or_zero(self.cx)
        cy = _length_or_zero(self.cy)
        rx = _length_or_zero(self.rx)
        ry = _length_or_zero(self.ry)

        return _ellipse_to_path_data(cx=cx, cy=cy, rx=rx, ry=ry)

    @override
    def to_path(self) -> Path:
        return _basic_shape_to_path(self)


@final
class FeBlend(
    attrdefs.ClassAttr,
    attrdefs.InAttr,
    attrdefs.In2Attr,
    attrdefs.ModeAttr,
    attrdefs.StyleAttr,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeColorMatrix(
    attrdefs.ClassAttr,
    attrdefs.InAttr,
    attrdefs.In2Attr,
    attrdefs.StyleAttr,
    attrdefs.TypeFeColorMatrixAttr,
    attrdefs.ValuesListOfNumbersAttr,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeComponentTransfer(
    attrdefs.ClassAttr,
    attrdefs.InAttr,
    attrdefs.StyleAttr,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeComposite(
    attrdefs.ClassAttr,
    attrdefs.InAttr,
    attrdefs.In2Attr,
    attrdefs.K1Attr,
    attrdefs.K2Attr,
    attrdefs.K3Attr,
    attrdefs.K4Attr,
    attrdefs.OperatorFeCompositeAttr,
    attrdefs.StyleAttr,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeConvolveMatrix(
    attrdefs.BiasAttr,
    attrdefs.ClassAttr,
    attrdefs.DivisorAttr,
    attrdefs.InAttr,
    attrdefs.KernelMatrixAttr,
    attrdefs.KernelUnitLengthAttr,
    attrdefs.OrderAttr,
    attrdefs.PreserveAlphaAttr,
    attrdefs.StyleAttr,
    attrdefs.TargetXAttr,
    attrdefs.TargetYAttr,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeDiffuseLighting(
    attrdefs.ClassAttr,
    attrdefs.DiffuseConstantAttr,
    attrdefs.InAttr,
    attrdefs.KernelUnitLengthAttr,
    attrdefs.StyleAttr,
    attrdefs.SurfaceScaleAttr,
    traits.Element,
):
    pass


@final
class FeDisplacementMap(
    attrdefs.ClassAttr,
    attrdefs.InAttr,
    attrdefs.In2Attr,
    attrdefs.ScaleAttr,
    attrdefs.StyleAttr,
    attrdefs.XChannelSelectorAttr,
    attrdefs.YChannelSelectorAttr,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeDistantLight(
    attrdefs.AzimuthAttr,
    attrdefs.ElevationAttr,
    traits.LightSourceElement,
    traits.Element,
):
    pass


@final
class FeFlood(
    attrdefs.ClassAttr,
    attrdefs.StyleAttr,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeFuncA(attrgroups.TransferFunctionAttrs, traits.Element):
    pass


@final
class FeFuncB(attrgroups.TransferFunctionAttrs, traits.Element):
    pass


@final
class FeFuncG(attrgroups.TransferFunctionAttrs, traits.Element):
    pass


@final
class FeFuncR(attrgroups.TransferFunctionAttrs, traits.Element):
    pass


@final
class FeGaussianBlur(
    attrdefs.ClassAttr,
    attrdefs.InAttr,
    attrdefs.StdDeviationAttr,
    attrdefs.StyleAttr,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeImage(
    attrgroups.XlinkAttrs,
    attrdefs.ClassAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.PreserveAspectRatioAttr,
    attrdefs.StyleAttr,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeMerge(
    attrdefs.ClassAttr,
    attrdefs.StyleAttr,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeMergeNode(attrdefs.InAttr, traits.Element):
    pass


@final
class FeMorphology(
    attrdefs.ClassAttr,
    attrdefs.InAttr,
    attrdefs.OperatorFeMorphologyAttr,
    attrdefs.RadiusAttr,
    attrdefs.StyleAttr,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeOffset(
    attrdefs.ClassAttr,
    attrdefs.DxNumberAttr,
    attrdefs.DyNumberAttr,
    attrdefs.InAttr,
    attrdefs.StyleAttr,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FePointLight(
    attrdefs.XNumberAttr,
    attrdefs.YNumberAttr,
    attrdefs.ZAttr,
    traits.LightSourceElement,
    traits.Element,
):
    pass


@final
class FeSpecularLighting(
    attrdefs.ClassAttr,
    attrdefs.InAttr,
    attrdefs.KernelUnitLengthAttr,
    attrdefs.SpecularConstantAttr,
    attrdefs.SpecularExponentAttr,
    attrdefs.StyleAttr,
    attrdefs.SurfaceScaleAttr,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeSpotLight(
    attrdefs.LimitingConeAngleAttr,
    attrdefs.PointsAtXAttr,
    attrdefs.PointsAtYAttr,
    attrdefs.PointsAtZAttr,
    attrdefs.SpecularExponentAttr,
    attrdefs.XNumberAttr,
    attrdefs.YNumberAttr,
    attrdefs.ZAttr,
    traits.LightSourceElement,
    traits.Element,
):
    pass


@final
class FeTile(
    attrdefs.ClassAttr,
    attrdefs.InAttr,
    attrdefs.StyleAttr,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class FeTurbulence(
    attrdefs.BaseFrequencyAttr,
    attrdefs.ClassAttr,
    attrdefs.NumOctavesAttr,
    attrdefs.SeedAttr,
    attrdefs.StitchTilesAttr,
    attrdefs.StyleAttr,
    attrdefs.TypeFeTurbluenceAttr,
    traits.FilterPrimitiveElement,
    traits.Element,
):
    pass


@final
class Filter(
    attrgroups.XlinkAttrs,
    attrdefs.ClassAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.FilterResAttr,
    attrdefs.FilterUnitsAttr,
    attrdefs.HeightAttr,
    attrdefs.PrimitiveUnitsAttr,
    attrdefs.StyleAttr,
    attrdefs.WidthAttr,
    attrdefs.XCoordinateAttr,
    attrdefs.YCoordinateAttr,
    traits.Element,
):
    pass


@final
class Font(
    attrdefs.ClassAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.HorizAdvXAttr,
    attrdefs.HorizOriginXAttr,
    attrdefs.HorizOriginYAttr,
    attrdefs.StyleAttr,
    attrdefs.VertAdvYAttr,
    attrdefs.VertOriginXAttr,
    attrdefs.VertOriginYAttr,
    traits.Element,
):
    pass


@final
class FontFace(
    attrdefs.AccentHeightAttr,
    attrdefs.AlphabeticAttr,
    attrdefs.AscentAttr,
    attrdefs.BboxAttr,
    attrdefs.CapHeightAttr,
    attrdefs.DescentAttr,
    attrdefs.HangingAttr,
    attrdefs.IdeographicAttr,
    attrdefs.MathematicalAttr,
    attrdefs.OverlinePositionAttr,
    attrdefs.OverlineThicknessAttr,
    attrdefs.Panose1Attr,
    attrdefs.SlopeAttr,
    attrdefs.StemhAttr,
    attrdefs.StemvAttr,
    attrdefs.StrikethroughPositionAttr,
    attrdefs.StrikethroughThicknessAttr,
    attrdefs.UnderlinePositionAttr,
    attrdefs.UnderlineThicknessAttr,
    attrdefs.UnicodeRangeAttr,
    attrdefs.UnitsPerEmAttr,
    attrdefs.VAlphabeticAttr,
    attrdefs.VHangingAttr,
    attrdefs.VIdeographicAttr,
    attrdefs.VMathematicalAttr,
    attrdefs.WidthsAttr,
    attrdefs.XHeightAttr,
    traits.Element,
):
    pass


@final
class FontFaceFormat(attrdefs.StringAttr, traits.Element):
    pass


@final
class FontFaceName(attrdefs.StringAttr, traits.Element):
    pass


@final
class FontFaceSrc(traits.Element):
    pass


@final
class FontFaceUri(attrgroups.XlinkAttrs, traits.Element):
    pass


@final
class ForeignObject(
    attrgroups.ConditionalProcessingAttrs,
    attrdefs.ClassAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.HeightAttr,
    attrdefs.StyleAttr,
    attrdefs.WidthAttr,
    attrdefs.XCoordinateAttr,
    attrdefs.YCoordinateAttr,
    traits.Element,
):
    pass


@final
class G(
    attrgroups.ConditionalProcessingAttrs,
    attrdefs.ClassAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.StyleAttr,
    traits.StructuralElement,
    traits.ContainerElement,
):
    pass


@final
class Glyph(
    attrdefs.ArabicFormAttr,
    attrdefs.ClassAttr,
    attrdefs.DAttr,
    attrdefs.GlyphNameAttr,
    attrdefs.HorizAdvXAttr,
    attrdefs.LangGlyphAttr,
    attrdefs.OrientationAttr,
    attrdefs.StyleAttr,
    attrdefs.UnicodeAttr,
    attrdefs.VertAdvYAttr,
    attrdefs.VertOriginXAttr,
    attrdefs.VertOriginYAttr,
    traits.ContainerElement,
):
    pass


@final
class GlyphRef(
    attrgroups.XlinkAttrs,
    attrdefs.ClassAttr,
    attrdefs.DxNumberAttr,
    attrdefs.DyNumberAttr,
    attrdefs.FormatAttr,
    attrdefs.GlyphRefAttr,
    attrdefs.StyleAttr,
    attrdefs.XNumberAttr,
    attrdefs.YNumberAttr,
    traits.Element,
):
    pass


@final
class Hkern(
    attrdefs.G1Attr,
    attrdefs.G2Attr,
    attrdefs.KAttr,
    attrdefs.U1Attr,
    attrdefs.U2Attr,
    traits.Element,
):
    pass


@final
class Image(
    attrgroups.ConditionalProcessingAttrs,
    attrgroups.XlinkAttrs,
    attrdefs.ClassAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.HeightAttr,
    attrdefs.PreserveAspectRatioAttr,
    attrdefs.StyleAttr,
    attrdefs.WidthAttr,
    attrdefs.XCoordinateAttr,
    attrdefs.YCoordinateAttr,
    traits.GraphicsElement,
    traits.GraphicsReferencingElement,
    traits.Element,
):
    pass


@final
class Line(
    attrgroups.ConditionalProcessingAttrs,
    attrdefs.ClassAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.StyleAttr,
    attrdefs.X1Attr,
    attrdefs.X2Attr,
    attrdefs.Y1Attr,
    attrdefs.Y2Attr,
    traits.BasicShape,
    traits.Element,
):
    @override
    def to_path_data(self) -> path_data.PathData:
        x1 = _length_or_zero(self.x1)
        y1 = _length_or_zero(self.y1)
        x2 = _length_or_zero(self.x2)
        y2 = _length_or_zero(self.y2)

        return (
            path_data.PathData()
            .move_to(point.Point(x1, y1))
            .line_to(point.Point(x2, y2))
        )

    @override
    def to_path(self) -> Path:
        return _basic_shape_to_path(self)


@final
class LinearGradient(
    attrgroups.XlinkAttrs,
    attrdefs.ClassAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.GradientTransformAttr,
    attrdefs.GradientUnitsAttr,
    attrdefs.SpreadMethodAttr,
    attrdefs.StyleAttr,
    attrdefs.X1Attr,
    attrdefs.X2Attr,
    attrdefs.Y1Attr,
    attrdefs.Y2Attr,
    traits.GradientElement,
    traits.Element,
):
    pass


@final
class Marker(
    attrdefs.ClassAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.MarkerHeightAttr,
    attrdefs.MarkerUnitsAttr,
    attrdefs.MarkerWidthAttr,
    attrdefs.OrientAttr,
    attrdefs.PreserveAspectRatioAttr,
    attrdefs.RefXAttr,
    attrdefs.RefYAttr,
    attrdefs.StyleAttr,
    attrdefs.ViewBoxAttr,
    traits.ContainerElement,
):
    pass


@final
class Mask(
    attrdefs.ClassAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.HeightAttr,
    attrdefs.MaskContentUnitsAttr,
    attrdefs.MaskUnitsAttr,
    attrdefs.StyleAttr,
    attrdefs.WidthAttr,
    attrdefs.XCoordinateAttr,
    attrdefs.YCoordinateAttr,
    traits.ContainerElement,
):
    pass


@final
class Metadata(traits.DescriptiveElement, traits.Element):
    pass


@final
class MissingGlyph(
    attrdefs.ClassAttr,
    attrdefs.DAttr,
    attrdefs.HorizAdvXAttr,
    attrdefs.StyleAttr,
    attrdefs.VertAdvYAttr,
    attrdefs.VertOriginXAttr,
    attrdefs.VertOriginYAttr,
    traits.ContainerElement,
):
    pass


@final
class Mpath(
    attrgroups.XlinkAttrs,
    attrdefs.ExternalResourcesRequiredAttr,
    traits.Element,
):
    pass


@final
class Pattern(
    attrgroups.XlinkAttrs,
    attrdefs.ClassAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.HeightAttr,
    attrdefs.PatternContentUnitsAttr,
    attrdefs.PatternTransformAttr,
    attrdefs.PatternUnitsAttr,
    attrdefs.PreserveAspectRatioAttr,
    attrdefs.StyleAttr,
    attrdefs.ViewBoxAttr,
    attrdefs.WidthAttr,
    attrdefs.XCoordinateAttr,
    attrdefs.YCoordinateAttr,
    traits.ContainerElement,
):
    pass


@final
class Polygon(
    attrgroups.ConditionalProcessingAttrs,
    attrdefs.ClassAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.PointsAttr,
    attrdefs.StyleAttr,
    traits.BasicShape,
    traits.Element,
):
    @override
    def to_path_data(self) -> path_data.PathData:
        return _points_to_path_data(self).close()

    @override
    def to_path(self) -> Path:
        return _basic_shape_to_path(self)


@final
class Polyline(
    attrgroups.ConditionalProcessingAttrs,
    attrdefs.ClassAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.PointsAttr,
    attrdefs.StyleAttr,
    traits.BasicShape,
    traits.Element,
):
    @override
    def to_path_data(self) -> path_data.PathData:
        return _points_to_path_data(self)

    @override
    def to_path(self) -> Path:
        return _basic_shape_to_path(self)


@final
class RadialGradient(
    attrgroups.XlinkAttrs,
    attrdefs.ClassAttr,
    attrdefs.CxAttr,
    attrdefs.CyAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.FrAttr,
    attrdefs.FxAttr,
    attrdefs.FyAttr,
    attrdefs.GradientTransformAttr,
    attrdefs.GradientUnitsAttr,
    attrdefs.RAttr,
    attrdefs.SpreadMethodAttr,
    attrdefs.StyleAttr,
    traits.GradientElement,
    traits.Element,
):
    pass


@final
class Rect(
    attrgroups.ConditionalProcessingAttrs,
    attrdefs.ClassAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.HeightAttr,
    attrdefs.RxAttr,
    attrdefs.RyAttr,
    attrdefs.StyleAttr,
    attrdefs.WidthAttr,
    attrdefs.XCoordinateAttr,
    attrdefs.YCoordinateAttr,
    traits.BasicShape,
    traits.Element,
):
    @override
    def to_path_data(self) -> path_data.PathData:
        x = _length_or_zero(self.x)
        y = _length_or_zero(self.y)
        width = _length_or_zero(self.width)
        height = _length_or_zero(self.height)

        match self.rx, self.ry:
            case None, None:
                rx = ry = length.Length(0)
            case length.Length(), None:
                rx = ry = self.rx
            case None, length.Length():
                rx = ry = self.ry
            case length.Length(), length.Length():
                rx = self.rx
                ry = self.ry

        return (
            path_data.PathData()
            .move_to(point.Point(x + rx, y))
            .horizontal_line_to(x + width - rx)
            .arc_to(
                point.Point(rx, ry),
                0,
                point.Point(x + width, y + ry),
                large=False,
                sweep=True,
            )
            .vertical_line_to(y + height - ry)
            .arc_to(
                point.Point(rx, ry),
                0,
                point.Point(x + width - rx, y + height),
                large=False,
                sweep=True,
            )
            .horizontal_line_to(x + rx)
            .arc_to(
                point.Point(rx, ry),
                0,
                point.Point(x, y + height - ry),
                large=False,
                sweep=True,
            )
            .vertical_line_to(y + ry)
            .arc_to(
                point.Point(rx, ry),
                0,
                point.Point(x + rx, y),
                large=False,
                sweep=True,
            )
        )

    @override
    def to_path(self) -> Path:
        return _basic_shape_to_path(self)


@final
class Script(
    attrgroups.XlinkAttrs,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.TypeContentTypeAttr,
    traits.Element,
):
    pass


@final
class Set(
    attrgroups.AnimationAttributeTargetAttrs,
    attrgroups.XlinkAttrs,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.ToAttr,
    traits.AnimationElement,
    traits.Element,
):
    pass


@final
class Stop(
    attrdefs.ClassAttr,
    attrdefs.OffsetNumberPercentageAttr,
    attrdefs.StyleAttr,
    traits.Element,
):
    pass


@final
class Style(
    attrdefs.MediaAttr,
    attrdefs.TitleAttr,
    attrdefs.TypeContentTypeAttr,
    traits.Element,
):
    pass


@final
class Svg(
    attrgroups.ConditionalProcessingAttrs,
    attrgroups.DocumentEventsAttrs,
    attrdefs.BaseProfileAttr,
    attrdefs.ClassAttr,
    attrdefs.ContentScriptTypeAttr,
    attrdefs.ContentStyleTypeAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.HeightAttr,
    attrdefs.PreserveAspectRatioAttr,
    attrdefs.StyleAttr,
    attrdefs.VersionAttr,
    attrdefs.ViewBoxAttr,
    attrdefs.WidthAttr,
    attrdefs.XCoordinateAttr,
    attrdefs.XmlnsAttr,
    attrdefs.YCoordinateAttr,
    attrdefs.ZoomAndPanAttr,
    traits.StructuralElement,
    traits.ContainerElement,
):
    """The `<svg>` element.

    The `<svg>` element is the root of an SVG document fragment.

    You can either create a new SVG document from scratch by creating an
    instance of this class, or you can parse an existing SVG document
    using the `parse_svg` function.
    """

    @overload
    def save(
        self,
        path: str | os.PathLike[str],
        /,
        *,
        pretty: bool = True,
        trailing_newline: bool = True,
        formatter: serialize.Formatter | None = None,
    ) -> None: ...

    @overload
    def save(
        self,
        file: protocols.SupportsWrite[str],
        /,
        *,
        pretty: bool = True,
        trailing_newline: bool = True,
        formatter: serialize.Formatter | None = None,
    ) -> None: ...

    def save(
        self,
        path_or_file: str
        | os.PathLike[str]
        | protocols.SupportsWrite[str],
        /,
        *,
        pretty: bool = True,
        trailing_newline: bool = True,
        formatter: serialize.Formatter | None = None,
    ) -> None:
        """Convert the SVG document fragment to XML and write it to a file.

        Args:
        path_or_file: The path to the file to save the XML to,
        or a file-like object.
        pretty: Whether to produce pretty-printed XML.
        indent: The number of spaces to indent each level of the document.
        trailing_newline: Whether to add a trailing newline to the file.
        formatter: The formatter to use for serialization.

        Examples:
        >>> import sys
        >>> from svglab import Rect, Svg
        >>> svg = Svg(id="foo").add_child(Rect())
        >>> formatter = serialize.Formatter(indent=4)
        >>> svg.save(
        ...     sys.stdout,
        ...     pretty=True,
        ...     trailing_newline=False,
        ...     formatter=formatter,
        ... )
        <svg id="foo">
            <rect/>
        </svg>

        """
        with contextlib.ExitStack() as stack:
            output = self.to_xml(pretty=pretty, formatter=formatter)
            file: protocols.SupportsWrite[str]

            match path_or_file:
                case str() | os.PathLike() as path:
                    file = stack.enter_context(
                        pathlib.Path(path).open("w")
                    )
                case protocols.SupportsWrite() as file:
                    pass

            file.write(output)

            if trailing_newline:
                file.write("\n")

    def set_viewbox(
        self, viewbox: tuple[float, float, float, float]
    ) -> None:
        """Set a new value for the `viewBox` attribute.

        This method sets a new value for the `viewBox` attribute and scales and
        translates the SVG content so that the visual representation of the SVG
        remains unchanged.

        If the `viewBox` is not set, the method uses the `width` and `height`
        attributes to calculate the initial viewBox. If the `width`
        and `height` attributes are not set, the method raises an exception.

        The new `viewBox` must have the same aspect ratio as the old `viewBox`.
        If the aspect ratios differ, the method raises an exception.

        Any attributes of type `Length` in the SVG must be convertible to
        user units. If an attribute is not convertible, the method raises an
        exception.

        Args:
        viewbox: A tuple of four numbers representing the new viewBox.

        Raises:
        ValueError: If `viewBox` is not set and `width` and `height` are not
            set or if the aspect ratios of the old and new viewBox differ.
        SvgUnitConversionError: If an attribute is not convertible to user
            units.

        """
        if self.viewBox is None:
            if self.width is None or self.height is None:
                raise ValueError(
                    "Either viewBox or width and height must be set"
                )
            old_viewbox = (0, 0, float(self.width), float(self.height))
        else:
            old_viewbox = self.viewBox

        old_min_x, old_min_y, old_width, old_height = old_viewbox
        min_x, min_y, width, height = viewbox

        tx = min_x - old_min_x
        ty = min_y - old_min_y

        sx = width / old_width
        sy = height / old_height

        if not mathutils.is_close(sx, sy):
            raise ValueError("Aspect ratios of old and new viewBox differ")

        # skip self; this can be done in a single for loop because the
        # SVG is a tree (probably)
        for child in self.find_all(recursive=False):
            # this is normally done in the reify method, but we need to do it
            # before we prepend the new transformations
            child.decompose_transform_origin()

            child.transform = [
                transform.Translate(tx, ty),
                transform.Scale(sx, sy),
                *(child.transform or []),
            ]

            child.reify(limit=2, recursive=False)

        self.viewBox = viewbox

    def render(
        self, *, width: float | None = None, height: float | None = None
    ) -> PIL.Image.Image:
        """Render an SVG document fragment into a Pillow image.

        If the width and height are not specified, the dimensions of the SVG
        element are used. If only one dimension is specified, the other
        dimension is calculated in a way that preserves the aspect ratio set
        in the SVG element. If both dimensions are specified, the SVG is
        scaled so that the aspect ratio is preserved.

        Args:
        svg: The SVG document fragment to render.
        width: The width of the rendered image, in pixels. If `None`, the width
            attribute of the SVG element is used.
        height: The height of the rendered image, in pixels. If `None`, the
            height attribute of the SVG element is used.

        Returns:
        The rendered image.

        """
        return graphics.render(self, width=width, height=height)

    def show(self) -> None:
        """Render this SVG document fragment and display it on the screen.

        See `PIL.Image.Image.show` for more information.

        """
        self.render().show()

    def to_data_uri(self) -> str:
        """Convert the SVG document fragment to a base64-encoded data URI."""
        xml = self.to_xml(formatter=serialize.MINIMAL_FORMATTER)
        b64 = base64.b64encode(xml.encode()).decode()

        return f"data:image/svg+xml;base64,{b64}"


@final
class Switch(
    attrgroups.ConditionalProcessingAttrs,
    attrdefs.ClassAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.StyleAttr,
    traits.ContainerElement,
):
    pass


@final
class Symbol(
    attrdefs.ClassAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.PreserveAspectRatioAttr,
    attrdefs.RefXAttr,
    attrdefs.RefYAttr,
    attrdefs.StyleAttr,
    attrdefs.ViewBoxAttr,
    traits.StructuralElement,
    traits.ContainerElement,
):
    pass


@final
class Text(
    attrgroups.ConditionalProcessingAttrs,
    attrdefs.ClassAttr,
    attrdefs.DxListOfLengthsAttr,
    attrdefs.DyListOfLengthsAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.LengthAdjustAttr,
    attrdefs.RotateListOfNumbersAttr,
    attrdefs.StyleAttr,
    attrdefs.TextLengthAttr,
    attrdefs.XListOfCoordinatesAttr,
    attrdefs.YListOfCoordinatesAttr,
    traits.TextContentElement,
    traits.Element,
):
    pass


@final
class TextPath(
    attrgroups.ConditionalProcessingAttrs,
    attrgroups.XlinkAttrs,
    attrdefs.ClassAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.MethodAttr,
    attrdefs.SpacingAttr,
    attrdefs.StartOffsetAttr,
    attrdefs.StyleAttr,
    traits.TextContentChildElement,
    traits.TextContentElement,
    traits.Element,
):
    pass


@final
class Title(
    attrdefs.ClassAttr,
    attrdefs.StyleAttr,
    traits.DescriptiveElement,
    traits.Element,
):
    pass


@final
class Tref(
    attrgroups.ConditionalProcessingAttrs,
    attrgroups.XlinkAttrs,
    attrdefs.ClassAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.StyleAttr,
    traits.TextContentChildElement,
    traits.TextContentElement,
    traits.Element,
):
    pass


@final
class Tspan(
    attrgroups.ConditionalProcessingAttrs,
    attrdefs.ClassAttr,
    attrdefs.DxListOfLengthsAttr,
    attrdefs.DyListOfLengthsAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.LengthAdjustAttr,
    attrdefs.RotateListOfNumbersAttr,
    attrdefs.StyleAttr,
    attrdefs.TextLengthAttr,
    attrdefs.XListOfCoordinatesAttr,
    attrdefs.YListOfCoordinatesAttr,
    traits.TextContentBlockElement,
    traits.TextContentChildElement,
    traits.TextContentElement,
    traits.Element,
):
    pass


@final
class Use(
    attrgroups.ConditionalProcessingAttrs,
    attrgroups.XlinkAttrs,
    attrdefs.ClassAttr,
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.HeightAttr,
    attrdefs.StyleAttr,
    attrdefs.WidthAttr,
    attrdefs.XCoordinateAttr,
    attrdefs.YCoordinateAttr,
    traits.GraphicsElement,
    traits.GraphicsReferencingElement,
    traits.StructuralElement,
    traits.Element,
):
    pass


@final
class View(
    attrdefs.ExternalResourcesRequiredAttr,
    attrdefs.PreserveAspectRatioAttr,
    attrdefs.ViewBoxAttr,
    attrdefs.ViewTargetAttr,
    attrdefs.ZoomAndPanAttr,
    traits.Element,
):
    pass


@final
class Vkern(
    attrdefs.G1Attr,
    attrdefs.G2Attr,
    attrdefs.KAttr,
    attrdefs.U1Attr,
    attrdefs.U2Attr,
    traits.Element,
):
    pass



"""Functions related to rendering and other graphics operations."""

import concurrent.futures
import copy
import io
import itertools
import uuid
from collections.abc import Iterable

import numpy as np
import numpy.typing as npt
import PIL.Image
import resvg_py
from typing_extensions import (
    Final,
    Protocol,
    TypeAlias,
    TypeVar,
    cast,
    runtime_checkable,
)

from svglab import entities, errors, serialize
from svglab.attrparse import color, length


Mask: TypeAlias = npt.NDArray[np.bool_]
BBox: TypeAlias = tuple[int, int, int, int]

_ImageArray: TypeAlias = npt.NDArray[np.uint8]

_ElementT = TypeVar("_ElementT", bound=entities.Element)

_BLACK: Final = color.Color((0, 0, 0))


@runtime_checkable
class _SvgElementLike(Protocol):
    width: length.Length | None
    height: length.Length | None

    def render(
        self, width: float | None = None, height: float | None = None
    ) -> PIL.Image.Image: ...


def _length_to_user_units(length: length.Length | None) -> float | None:
    """Convert a length to user units, if possible.

    Args:
        length: The length to convert, or `None`.

    Returns:
        The length in user units, or `None` if the length cannot be converted.

    """
    if length is None:
        return None

    try:
        return float(length)
    except errors.SvgUnitConversionError:
        return None


def _compute_render_size(
    svg: entities.Element,
    *,
    width: float | None = None,
    height: float | None = None,
) -> tuple[float, float]:
    if not isinstance(svg, _SvgElementLike):
        msg = "Svg must be an instance of _SvgElementLike"
        raise TypeError(msg)

    svg_width = _length_to_user_units(length=svg.width)
    svg_height = _length_to_user_units(svg.height)

    if svg_width is not None and svg_height is not None:
        if width is not None and height is None:
            ratio = width / svg_width
            height = svg_height * ratio
        elif width is None and height is not None:
            ratio = height / svg_height
            width = svg_width * ratio
        elif width is not None and height is not None:
            ratio = (svg_width / svg_height) / (width / height)

            if ratio > 1:
                height = height / ratio
            elif ratio < 1:
                width = width / ratio

    width = width if width is not None else svg_width
    height = height if height is not None else svg_height

    if width is None or height is None:
        msg = (
            "Unable to determine image dimensions: "
            f"{svg.width=}, {svg.height=}, {width=}, {height=}"
        )
        raise ValueError(msg)

    return width, height


def _resvg_render(markup: str) -> bytes:
    with concurrent.futures.ProcessPoolExecutor(max_workers=1) as executor:
        future = executor.submit(resvg_py.svg_to_bytes, markup)

        return cast(bytes, future.result())


def render(  # noqa: D103
    svg: entities.Element,
    *,
    width: float | None = None,
    height: float | None = None,
) -> PIL.Image.Image:
    if not isinstance(svg, _SvgElementLike):
        raise TypeError("Element must be an SVG element")

    render_size = _compute_render_size(svg, width=width, height=height)
    svg = copy.copy(svg)

    svg.width = length.Length(render_size[0])
    svg.height = length.Length(render_size[1])

    xml = svg.to_xml(formatter=serialize.MINIMAL_FORMATTER)
    raw = _resvg_render(xml)

    return PIL.Image.open(io.BytesIO(raw))


def _copy_tree(element: _ElementT) -> tuple[_ElementT, _SvgElementLike]:
    """Resolve the root `Svg` element and create a deep copy of the SVG tree.

    The source element is identified in the copied tree and returned for easy
    access.

    Args:
        element: The element in the SVG tree to copy.

    Returns:
        A tuple of the copied element and the root `Svg` element of the copied
        tree.

    Raises:
        ValueError: If the element is not a part of an SVG tree.

    """
    svg = element.get_root()

    if not isinstance(svg, _SvgElementLike):
        raise ValueError("Element must be part of an SVG tree")  # noqa: TRY004

    original_id = element.id
    element.id = uuid.uuid4().hex

    try:
        svg = copy.deepcopy(svg)
        candidates = cast(
            Iterable[_ElementT],
            itertools.chain([svg], svg.find_all(type(element))),
        )
        this = next(
            element for element in candidates if element.id == element.id
        )
    finally:
        element.id = original_id

    return this, svg


def _make_element_visible(element: entities.Element, /) -> None:
    del element.display
    element.fill = _BLACK
    element.fill_opacity = 1
    element.opacity = 1
    element.stroke = _BLACK
    element.stroke_opacity = 1
    element.visibility = "visible"

    for child in element.find_all(recursive=False):
        _make_element_visible(child)


def _render_tree(
    element: entities.Element,
    *,
    render_this: bool,
    render_other: bool,
    make_element_visible: bool,
    width: float | None = None,
    height: float | None = None,
) -> PIL.Image.Image:
    """Resolve the root `Svg` element and render the SVG tree to an image.

    Args:
        element: The element in the SVG tree to render.
        render_this: Whether to render the specified element.
        render_other: Whether to render all other elements in the tree.
        make_element_visible: Whether to attempt to make the specified element
            visible, even if it would normally not be rendered (e.g., if due
            to a transparent fill).
        width: The width of the rendered image, in pixels. If `None`, the width
            attribute of the SVG element is used.
        height: The height of the rendered image, in pixels. If `None`, the
            height attribute of the SVG element is used.

    Returns:
        The rendered image.

    Raises:
        ValueError: If `make_element_visible` is `True` and `render_this`
        is `False`.

    """
    if make_element_visible and not render_this:
        raise ValueError(
            "make_element_visible cannot be True if render_this is False"
        )

    element_copy, svg = _copy_tree(element)
    assert isinstance(svg, entities.Element)

    for t in svg.find_all():
        # do not hide the parents of our element as that would make it
        # invisible
        # TODO: this is quite suboptimal performance-wise; optimize this
        if render_other or t in element_copy.ancestors:
            t.visibility = "visible"
        else:
            t.visibility = "hidden"

    element_copy.visibility = "visible" if render_this else "hidden"

    if make_element_visible:
        _make_element_visible(element_copy)

    return svg.render(width=width, height=height)


def _mask_to_image(mask: Mask) -> PIL.Image.Image:
    """Convert boolean mask into an RGBA image.

    Areas where the mask is True are set to solid black. All other areas are
    fully transparent. This allows convenient use with
    `PIL.Image.Image.getbbox()`.

    Args:
        mask: A boolean mask given as an NDArray of shape (x, y), where x
            and y are the dimensions of the resulting image.

    Returns:
        An RGBA image representing the mask.

    """
    x, y = mask.shape

    # create fully transparent image
    rgba: _ImageArray = np.zeros((x, y, 4), dtype=np.uint8)

    # set to solid black where mask is True
    rgba[mask, 3] = 255

    return PIL.Image.fromarray(rgba)


def mask(  # noqa: D103
    element: entities.Element,
    *,
    width: float | None = None,
    height: float | None = None,
) -> Mask:
    img = _render_tree(
        element,
        render_this=True,
        render_other=False,
        make_element_visible=True,
        width=width,
        height=height,
    )
    array: Mask = np.array(img)

    return array[:, :, 3] > 0  # alpha channel > 0


def visible_mask(  # noqa: D103
    element: entities.Element,
    *,
    width: float | None = None,
    height: float | None = None,
) -> Mask:
    without_element = _render_tree(
        element,
        render_this=False,
        render_other=True,
        make_element_visible=False,
        width=width,
        height=height,
    )

    with_element = _render_tree(
        element,
        render_this=True,
        render_other=True,
        make_element_visible=False,
        width=width,
        height=height,
    )

    without_element_array: _ImageArray = np.array(without_element)
    with_element_array: _ImageArray = np.array(with_element)

    diff = np.any(without_element_array != with_element_array, axis=2)
    assert isinstance(diff, np.ndarray)

    return diff


def bbox(element: entities.Element) -> BBox | None:  # noqa: D103
    img = _render_tree(
        element,
        render_this=True,
        render_other=False,
        make_element_visible=True,
    )

    return img.getbbox()


def visible_bbox(element: entities.Element) -> BBox | None:  # noqa: D103
    mask = visible_mask(element)
    img = _mask_to_image(mask)

    return img.getbbox()



"""Names of all elements.

This module contains the names of all elements in the SVG specification.
It also provides functions to normalize these names to Python identifiers.
"""

import bidict
import typing_extensions
from typing_extensions import Final, Literal, TypeAlias


ElementName: TypeAlias = Literal[
    "a",
    "altGlyph",
    "altGlyphDef",
    "altGlyphItem",
    "animate",
    "animateColor",
    "animateMotion",
    "animateTransform",
    "circle",
    "clipPath",
    "color-profile",
    "cursor",
    "defs",
    "desc",
    "ellipse",
    "feBlend",
    "feColorMatrix",
    "feComponentTransfer",
    "feComposite",
    "feConvolveMatrix",
    "feDiffuseLighting",
    "feDisplacementMap",
    "feDistantLight",
    "feFlood",
    "feFuncA",
    "feFuncB",
    "feFuncG",
    "feFuncR",
    "feGaussianBlur",
    "feImage",
    "feMerge",
    "feMergeNode",
    "feMorphology",
    "feOffset",
    "fePointLight",
    "feSpecularLighting",
    "feSpotLight",
    "feTile",
    "feTurbulence",
    "filter",
    "font-face-format",
    "font-face-name",
    "font-face-src",
    "font-face-uri",
    "font-face",
    "font",
    "foreignObject",
    "g",
    "glyph",
    "glyphRef",
    "hkern",
    "image",
    "line",
    "linearGradient",
    "marker",
    "mask",
    "metadata",
    "missing-glyph",
    "mpath",
    "path",
    "pattern",
    "polygon",
    "polyline",
    "radialGradient",
    "rect",
    "script",
    "set",
    "stop",
    "style",
    "svg",
    "switch",
    "symbol",
    "text",
    "textPath",
    "title",
    "tref",
    "tspan",
    "use",
    "view",
    "vkern",
]
"""Type for all SVG element names."""


ELEMENT_NAMES: Final[frozenset[ElementName]] = frozenset(
    typing_extensions.get_args(ElementName)
)
"""A set of all SVG element names."""


def _normalize_element_name(name: ElementName, /) -> str:
    """Convert an SVG element name to an appropriate class name.

    Args:
        name: The element name to normalize.

    Returns:
        The normalized element name in `PascalCase`.

    Raises:
        ValueError: If the element name cannot be normalized.

    Examples:
    >>> _normalize_element_name("circle")
    'Circle'
    >>> _normalize_element_name("feGaussianBlur")
    'FeGaussianBlur'
    >>> _normalize_element_name("font-face-name")
    'FontFaceName'

    """
    return "".join(part[0].upper() + part[1:] for part in name.split("-"))


ELEMENT_NAME_TO_NORMALIZED: Final = bidict.frozenbidict[ElementName, str](
    {
        element: _normalize_element_name(element)
        for element in ELEMENT_NAMES
    }
)
"""
A bidirectional mapping from SVG element names to normalized Python
identifiers.
"""



"""Groups of attributes that are commonly used together."""

from svglab.attrs import attrdefs


class CoreAttrs(
    attrdefs.IdAttr,
    attrdefs.LangAttr,
    attrdefs.XmlBaseAttr,
    attrdefs.XmlLangAttr,
    attrdefs.XmlSpaceAttr,
):
    """The core attributes.

    From the SVG 1.1 specification:
    > "The core attributes are those attributes that can be specified on any
    SVG element."
    """


class PresentationAttrs(
    attrdefs.AlignmentBaselineAttr,
    attrdefs.BaselineShiftAttr,
    attrdefs.ClipAttr,
    attrdefs.ClipPathAttr,
    attrdefs.ClipRuleAttr,
    attrdefs.ColorAttr,
    attrdefs.ColorInterpolationAttr,
    attrdefs.ColorInterpolationFiltersAttr,
    attrdefs.ColorProfileAttr,
    attrdefs.ColorRenderingAttr,
    attrdefs.CursorAttr,
    attrdefs.DirectionAttr,
    attrdefs.DisplayAttr,
    attrdefs.DominantBaselineAttr,
    attrdefs.EnableBackgroundAttr,
    attrdefs.FillAttr,
    attrdefs.FillOpacityAttr,
    attrdefs.FillRuleAttr,
    attrdefs.FilterAttr,
    attrdefs.FloodColorAttr,
    attrdefs.FloodOpacityAttr,
    attrdefs.FontFamilyAttr,
    attrdefs.FontSizeAdjustAttr,
    attrdefs.FontSizeAttr,
    attrdefs.FontStretchAttr,
    attrdefs.FontStyleAttr,
    attrdefs.FontVariantAttr,
    attrdefs.FontWeightAttr,
    attrdefs.GlyphOrientationHorizontalAttr,
    attrdefs.GlyphOrientationVerticalAttr,
    attrdefs.ImageRenderingAttr,
    attrdefs.KerningAttr,
    attrdefs.LetterSpacingAttr,
    attrdefs.LightingColorAttr,
    attrdefs.MarkerEndAttr,
    attrdefs.MarkerMidAttr,
    attrdefs.MarkerStartAttr,
    attrdefs.MaskAttr,
    attrdefs.OpacityAttr,
    attrdefs.OverflowAttr,
    attrdefs.PaintOrderAttr,
    attrdefs.PointerEventsAttr,
    attrdefs.ShapeRenderingAttr,
    attrdefs.StopColorAttr,
    attrdefs.StopOpacityAttr,
    attrdefs.StrokeAttr,
    attrdefs.StrokeDasharrayAttr,
    attrdefs.StrokeDashoffsetAttr,
    attrdefs.StrokeLinecapAttr,
    attrdefs.StrokeLinejoinAttr,
    attrdefs.StrokeMiterlimitAttr,
    attrdefs.StrokeOpacityAttr,
    attrdefs.StrokeWidthAttr,
    attrdefs.TextAlignAllAttr,
    attrdefs.TextAlignAttr,
    attrdefs.TextAlignLastAttr,
    attrdefs.TextAnchorAttr,
    attrdefs.TextDecorationAttr,
    attrdefs.TextIndentAttr,
    attrdefs.TextRenderingAttr,
    attrdefs.TransformAttr,
    attrdefs.TransformOriginAttr,
    attrdefs.UnicodeBidiAttr,
    attrdefs.VectorEffectAttr,
    attrdefs.VisibilityAttr,
    attrdefs.WhiteSpaceAttr,
    attrdefs.WordSpacingAttr,
    attrdefs.WritingModeAttr,
):
    """The presentation attributes.

    From the SVG 1.1 specification:
    > "An XML attribute on an SVG element which specifies a value for a given
    property for that element."
    """


class AnimationEventsAttrs(
    attrdefs.OnBeginAttr,
    attrdefs.OnEndAttr,
    attrdefs.OnLoadAttr,
    attrdefs.OnRepeatAttr,
):
    """The animation event attributes.

    From the SVG 1.1 specification:
    > "An animation event attribute is an event attribute that specifies script
    to run for a particular animation-related event."
    """


class AnimationAttributeTargetAttrs(
    attrdefs.AttributeNameAttr, attrdefs.AttributeTypeAttr
):
    """The animation attribute target attributes.

    From the SVG 1.1 specification:
    > "[...] identify the target attribute or property for the given target
    element whose value changes over time."

    """


class AnimationTimingAttrs(
    attrdefs.BeginAttr,
    attrdefs.DurAttr,
    attrdefs.EndAttr,
    attrdefs.FillAttr,
    attrdefs.MaxAttr,
    attrdefs.MinAttr,
    attrdefs.RepeatCountAttr,
    attrdefs.RepeatDurAttr,
    attrdefs.RestartAttr,
):
    """The animation timing attributes.

    From the SVG 1.1 specification:
    > "They are common to all animation elements and control the timing of the
    animation, including what causes the animation to start and end, whether
    the animation runs repeatedly, and whether to retain the end state the
    animation once the animation ends."
    """


class AnimationValueAttrs(
    attrdefs.ByAttr,
    attrdefs.CalcModeAttr,
    attrdefs.FromAttr,
    attrdefs.KeySplinesAttr,
    attrdefs.KeyTimesAttr,
    attrdefs.ToAttr,
    attrdefs.ValuesListAttr,
):
    """The animation value attributes.

    From the SVG 1.1 specification:
    > "They are common to elements `animate`, `animateColor`, `animateMotion`
    and `animateTransform`. These attributes define the values that are
    assigned to the target attribute or property over time. The attributes
    [...] provide control over the relative timing of keyframes and the
    interpolation method between discrete values."
    """


class AnimationAdditionAttrs(
    attrdefs.AccumulateAttr, attrdefs.AdditiveAttr
):
    """The animation addition attributes.

    From the SVG 1.1 specification:
    > "[...] are common to elements `animate`, `animateColor`, `animateMotion`
    and `animateTransform`."
    """


class ConditionalProcessingAttrs(
    attrdefs.RequiredFeaturesAttr,
    attrdefs.RequiredExtensionsAttr,
    attrdefs.SystemLanguageAttr,
):
    """The conditional processing attributes.

    From the SVG 1.1 specification:
    > "A conditional processing attribute is one that controls whether or not
    the element on which it appears is processed. Most elements, but not all,
    may have conditional processing attributes specified on them."
    """


class DocumentEventsAttrs(
    attrdefs.OnAbortAttr,
    attrdefs.OnErrorAttr,
    attrdefs.OnResizeAttr,
    attrdefs.OnScrollAttr,
    attrdefs.OnUnloadAttr,
    attrdefs.OnZoomAttr,
):
    """The document event attributes.

    From the SVG 1.1 specification:
    > "A document event attribute is an event attribute that specifies script
    to run for a particular document-wide event."
    """


class GraphicalEventsAttrs(
    attrdefs.OnActivateAttr,
    attrdefs.OnClickAttr,
    attrdefs.OnFocusInAttr,
    attrdefs.OnFocusOutAttr,
    attrdefs.OnLoadAttr,
    attrdefs.OnMouseDownAttr,
    attrdefs.OnMouseMoveAttr,
    attrdefs.OnMouseOutAttr,
    attrdefs.OnMouseOverAttr,
    attrdefs.OnMouseUpAttr,
):
    """The graphical event attributes.

    From the SVG 1.1 specification:
    > "A graphical event attribute is an event attribute that specifies script
    to run for a particular user interaction event."
    """


class EventAttrs(
    GraphicalEventsAttrs, AnimationEventsAttrs, DocumentEventsAttrs
):
    """All event attributes."""


class FilterPrimitivesAttrs(
    attrdefs.XCoordinateAttr,
    attrdefs.YCoordinateAttr,
    attrdefs.WidthAttr,
    attrdefs.HeightAttr,
    attrdefs.ResultAttr,
):
    """The filter primitive attributes.

    From the SVG 1.1 specification:
    > "The filter primitive attributes is the set of attributes that are common
    to all filter primitive elements."
    """


class XlinkAttrs(
    attrdefs.HrefAttr,
    attrdefs.XlinkActuateOnLoadAttr,
    attrdefs.XlinkArcroleAttr,
    attrdefs.XlinkHrefAttr,
    attrdefs.XlinkRoleAttr,
    attrdefs.XlinkShowAttr,
    attrdefs.XlinkTitleAttr,
    attrdefs.XlinkTypeAttr,
):
    """The XLink attributes.

    From the SVG 1.1 specification:
    > "The XLink attributes are the seven attributes defined in the XML Linking
    Language specification, which are used on various SVG elements that can
    reference resources."
    """


class TransferFunctionAttrs(
    attrdefs.AmplitudeAttr,
    attrdefs.ExponentAttr,
    attrdefs.InterceptAttr,
    attrdefs.OffsetNumberAttr,
    attrdefs.SlopeAttr,
    attrdefs.TableValuesAttr,
    attrdefs.TypeFeFuncAttr,
):
    """The transfer function attributes.

    From the SVG 1.1 specification:
    > "[...] apply to sub-elements `feFuncR`, `feFuncG`, `feFuncB` and
    `feFuncA` that define the transfer functions."
    """
